#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import abc
import glob
import itertools
import json
import logging
import os
from collections.abc import Mapping

import toml
import yaml

from kapitan import cached
from kapitan.errors import CompileError, KapitanError
from kapitan.inventory.model.input_types import CompileInputTypeConfig, OutputType
from kapitan.refs.base import Revealer
from kapitan.utils import PrettyDumper, prune_empty

logger = logging.getLogger(__name__)


class InputType(object):
    """
    Abstract base class for input types.

    Provides methods for compiling input files.  Subclasses should implement
    the `compile_file` method to handle specific input formats.

    """

    __metaclass__ = abc.ABCMeta
    output_type_default = OutputType.YAML

    def __init__(self, compile_path, search_paths, ref_controller, target_name, args):
        self.compile_path = compile_path
        self.search_paths = search_paths
        self.ref_controller = ref_controller
        self.target_name = target_name
        self.args = args

    def compile_obj(self, comp_obj: CompileInputTypeConfig):
        """Expand globbed input paths and compile each resolved input path.

        Args:
            comp_obj: CompileInputTypeConfig object containing input paths and other compilation options.

        Raises:
            CompileError: If an input path is not found and ignore_missing is False.

        """

        # expand any globbed paths, taking into account provided search paths
        expanded_input_paths = []
        for input_path in comp_obj.input_paths:
            globbed_paths = [glob.glob(os.path.join(path, input_path)) for path in self.search_paths]
            inputs = list(itertools.chain.from_iterable(globbed_paths))
            # remove duplicate inputs
            inputs = set(inputs)
            ignore_missing = comp_obj.ignore_missing
            if len(inputs) == 0 and not ignore_missing:
                raise CompileError(
                    "Compile error: {} for target: {} not found in "
                    "search_paths: {}".format(input_path, self.target_name, self.search_paths)
                )
            expanded_input_paths.extend(inputs)

        for expanded_path in expanded_input_paths:
            self.compile_input_path(comp_obj, expanded_path)

    def to_file(self, config: CompileInputTypeConfig, file_path, file_content):
        """Write compiled content to file, handling different output types and revealing refs if needed.

        Args:
            config: CompileInputTypeConfig object.
            file_path: Path to the output file.
            file_content: Compiled content to write.

        Raises:
            ValueError: If the output type is not supported.
        """
        reveal = self.args.reveal
        target_name = self.target_name

        indent = self.args.indent
        output_type = config.output_type
        file_ext = output_type

        if config.prune:
            file_content = prune_empty(file_content)

        if output_type == OutputType.AUTO:
            _, detected_type = os.path.splitext(file_path)
            if detected_type:
                # Remove . from the beginning of the extension
                detected_type = detected_type[1:]

            if detected_type in [OutputType.TOML, OutputType.JSON, OutputType.YAML, OutputType.YML]:
                output_type = detected_type
                file_ext = None
            else:
                # Extension is not handled, falling back to input type default
                output_type = self.output_type_default
                logger.debug("Could not detect extension for %s, defaulting to %s", file_path, output_type)
                file_ext = output_type  # no extension for plain text

        if output_type == OutputType.PLAIN:
            file_ext = None  # no extension for plain text

        file_name = f"{file_path}.{file_ext}" if file_ext else file_path

        with CompiledFile(
            # file_path: path to the output file
            file_name,
            # ref_controller: reference controller to resolve refs
            self.ref_controller,
            # mode: file open mode, 'w' for write
            mode="w",
            # reveal: reveal refs in output
            reveal=reveal,
            target_name=target_name,
            indent=indent,
        ) as fp:
            if output_type == OutputType.JSON:
                fp.write_json(file_content)
            elif output_type in [OutputType.YAML, OutputType.YML]:
                fp.write_yaml(file_content)
            elif output_type == OutputType.PLAIN:
                fp.write(file_content)
            elif output_type == OutputType.TOML:
                fp.write_toml(file_content)
            else:
                raise ValueError(
                    f"Output type defined in inventory for {config} not supported: {output_type}: {OutputType}"
                )

    def compile_input_path(self, comp_obj: CompileInputTypeConfig, input_path: str):
        """Compile a single input path and write the result to the output directory.

        Creates the output directory if it doesn't exist.

        Args:
            comp_obj: CompileInputTypeConfig object.
            input_path: Path to the input file.
        Raises:
            CompileError: If compilation fails.

        """
        target_name = self.target_name
        output_path = comp_obj.output_path

        logger.debug("Compiling %s", input_path)

        try:
            target_compile_path = os.path.join(self.compile_path, target_name.replace(".", "/"), output_path)
            os.makedirs(target_compile_path, exist_ok=True)

            self.compile_file(
                comp_obj,
                input_path,
                target_compile_path,
            )

        except KapitanError as e:
            raise CompileError(
                "{}\nCompile error: failed to compile target: {}".format(e, target_name)
            ) from e

    @abc.abstractmethod
    def compile_file(self, config: CompileInputTypeConfig, input_path: str, compile_path: str):
        """Compile a single input file.

        This is an abstract method that should be implemented by subclasses to handle
        specific input formats.

        Args:
            config: CompileInputTypeConfig object.
            input_path: Path to the input file.
            compile_path: Path to the output directory.
        """
        return NotImplementedError


class CompilingFile(object):
    """
    A class to handle writing compiled data to a file.

    Provides methods to write data in different formats (YAML, JSON, TOML) to a file,
    handling reference revealing and compilation as needed.

    Args:
        fp: File object to write to.
        ref_controller: Reference controller to resolve references.
        **kwargs: Additional keyword arguments (e.g., reveal, target_name, indent).
    """

    def __init__(self, fp, ref_controller, **kwargs):
        self.fp = fp
        self.kwargs = kwargs
        self.revealer = Revealer(ref_controller)

    def write(self, data):
        """Write data into file.

        Args:
            data: Data to write.

        Reveals references if `reveal` is True in kwargs.

        """
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)

        if reveal:
            self.fp.write(self.revealer.reveal_raw(data))
        else:
            self.fp.write(self.revealer.compile_raw(data, target_name=target_name))

    def write_yaml(self, obj: object):
        """Recursively compile or reveal refs and convert obj to YAML and write to file.

        Args:
            obj: Object to write.

        Uses PrettyDumper to handle multiline strings according to style selection.
        """
        indent = self.kwargs.get("indent", 2)
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)

        # TODO(ademaria): make it configurable per input type
        style_selection = cached.inv[target_name]["parameters"].get("multiline_string_style", None)

        if not style_selection:
            if hasattr(cached.args, "multiline_string_style"):
                style_selection = cached.args.multiline_string_style
            elif hasattr(cached.args, "yaml_multiline_string_style"):
                style_selection = cached.args.yaml_multiline_string_style

        dumper = PrettyDumper.get_dumper_for_style(style_selection)

        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)

        if obj:
            if isinstance(obj, Mapping):
                yaml.dump(
                    obj,
                    stream=self.fp,
                    indent=indent,
                    Dumper=dumper,
                    default_flow_style=False,
                )
            else:
                yaml.dump_all(
                    obj,
                    stream=self.fp,
                    indent=indent,
                    Dumper=dumper,
                    default_flow_style=False,
                )

            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)

    def write_json(self, obj: object):
        """Recursively compile or reveal refs and convert obj to JSON and write to file.

        Args:
            obj: Object to write.

        Reveals references if `reveal` is True in kwargs.
        """
        indent = self.kwargs.get("indent", 2)
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)
        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)

        if obj:
            json.dump(obj, self.fp, indent=indent)
            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)

    def write_toml(self, obj: object):
        """Recursively compile or reveal refs and convert obj to TOML and write to file.

        Args:
            obj: Object to write.

        Reveals references if `reveal` is True in kwargs.
        """
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)
        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)
        if obj:
            toml.dump(obj, self.fp)
            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)


class CompiledFile(object):
    """
    A context manager to handle writing compiled data to a file.

    Args:
        name: Name of the output file.
        ref_controller: Reference controller to resolve references.
        **kwargs: Additional keyword arguments (e.g., mode).

    """

    def __init__(self, name, ref_controller, **kwargs):
        self.name = name
        self.fp = None
        self.ref_controller = ref_controller
        self.kwargs = kwargs

    def __enter__(self):
        mode = self.kwargs.get("mode", "r")

        # make sure directory for file exists
        os.makedirs(os.path.dirname(self.name), exist_ok=True)

        self.fp = open(self.name, mode)
        return CompilingFile(self.fp, self.ref_controller, **self.kwargs)

    def __exit__(self, *args):
        self.fp.close()



from typing import Annotated, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from kapitan.utils import StrEnum


class KapitanDependencyTypes(StrEnum):
    HELM = "helm"
    HTTP = "http"
    HTTPS = "https"
    GIT = "git"


class KapitanDependencyBaseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: KapitanDependencyTypes
    source: str
    output_path: str
    force_fetch: Optional[bool] = False


class KapitanDependencyHelmConfig(KapitanDependencyBaseConfig):
    type: Literal[KapitanDependencyTypes.HELM] = KapitanDependencyTypes.HELM
    chart_name: str
    version: Optional[str] = None
    helm_path: Optional[str] = None


class KapitanDependencyGitConfig(KapitanDependencyBaseConfig):
    type: Literal[KapitanDependencyTypes.GIT] = KapitanDependencyTypes.GIT
    ref: Optional[str] = "master"
    subdir: Optional[str] = None
    submodules: Optional[bool] = False


class KapitanDependencyHttpsConfig(KapitanDependencyBaseConfig):
    type: Literal[KapitanDependencyTypes.HTTPS, KapitanDependencyTypes.HTTP]
    unpack: bool = False


DependencyTypeConfig = Annotated[
    Union[KapitanDependencyHelmConfig, KapitanDependencyHttpsConfig, KapitanDependencyGitConfig],
    Field(discriminator="type"),
]



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"hashicorp vault kv secrets module"

import base64
import logging

from hvac.exceptions import Forbidden, InvalidPath

from kapitan import cached
from kapitan.inventory.model.references import (
    KapitanReferenceVaultKVConfig,
    VaultEngineTypes,
)
from kapitan.refs import KapitanReferencesTypes
from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend
from kapitan.refs.vault_resources import VaultClient, VaultError

logger = logging.getLogger(__name__)


class VaultSecret(Base64Ref):
    """
    Hashicorp Vault support for KV Secret Engine
    """

    def __init__(
        self,
        data,
        vault_params: KapitanReferenceVaultKVConfig,
        encrypt: bool = True,
        encode_base64=False,
        **kwargs,
    ):
        """
        Set vault parameter and encoding of data
        data will be passed as bytes
        """

        self.vault_params = vault_params

        if isinstance(vault_params, dict):
            self.vault_params = KapitanReferenceVaultKVConfig(**vault_params)

        if encrypt:
            # not really encryption --> storing key/value in vault
            self.mount = kwargs.get("mount_in_vault")
            self.path = kwargs.get("path_in_vault")
            self.key = kwargs.get("key_in_vault")
            self._encrypt(data, encode_base64)
        else:
            self.data = data

        super().__init__(self.data, **kwargs)
        self.type_name = KapitanReferencesTypes.VAULTKV

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new VaultSecret from data and ref_params: target_name, token
        parameters will be grabbed from the inventory via target_name
        vault parameters will be read from token
        """

        encoding = ref_params.kwargs.get("encoding", "original")
        if encoding == "original":
            data = data.encode()

        # set vault params as ref params
        if ref_params.kwargs.get("vault_params") is None:
            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv.get_parameters(target_name)

            try:
                vault_params = target_inv.kapitan.secrets.vaultkv
                ref_params.kwargs["vault_params"] = vault_params
            except KeyError:
                raise RefError("Could not create VaultSecret: vaultkv parameters missing")

        # set mount, path and key as ref params, read from token
        token = ref_params.kwargs.get("token")
        if token is None:
            raise RefError("Could not create VaultSecret: vaultkv parameters missing")

        token_attrs = token.split(":")

        if len(token_attrs) != 5:
            raise RefError("Could not create VaultSecret: ref token is invalid")

        # set mount
        mount = token_attrs[2]
        if not mount:
            mount = vault_params.mount
        ref_params.kwargs["mount_in_vault"] = mount

        # set path in vault
        path_in_vault = token_attrs[3]
        if not path_in_vault:
            path_in_vault = token_attrs[1]  # ref path in kapitan as default
        ref_params.kwargs["path_in_vault"] = path_in_vault

        # set key
        key = token_attrs[4]
        if key:
            ref_params.kwargs["key_in_vault"] = token_attrs[4]
        else:
            raise RefError("Could not create VaultSecret: vaultkv: key is missing")

        return cls(data, **ref_params.kwargs)

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        Returns decrypted data
        """
        # can't use super().reveal() as we want bytes
        if self.encoding == "base64":
            self.data = self.data.encode()

        return self._decrypt()

    def _encrypt(self, data, encode_base64=False):
        """
        Authenticate with Vault server & write given data in path/key
        """
        if self.path is None:
            raise VaultError(
                "Invalid path: None, you have to specify the path where the secret gets stored (in vault)"
            )

        # get vault client
        client = VaultClient(self.vault_params)
        secrets = {}

        # fetch current secrets from vault
        try:
            if self.vault_params.engine == VaultEngineTypes.KV:
                response = client.secrets.kv.v1.read_secret(
                    path=self.path,
                    mount_point=self.mount,
                )
                secrets = response["data"]
            else:
                response = client.secrets.kv.v2.read_secret_version(
                    path=self.path,
                    mount_point=self.mount,
                    raise_on_deleted_version=False,
                )
                secrets = response["data"]["data"]
        except InvalidPath:
            # path doesn't exist, will create new secret
            pass

        # append new secret
        secrets[self.key] = data.decode()

        # write updated secrets back to vault
        try:
            client.secrets.kv.v2.create_or_update_secret(
                path=self.path, secret=secrets, mount_point=self.mount
            )
            client.adapter.close()
        except Forbidden:
            raise VaultError(
                "Permission Denied. "
                + "make sure the token is authorised to access '{}' on Vault".format(self.path)
            )

        # set the data to path:key
        data = f"{self.path}:{self.key}".encode()

        self.encoding = "original"
        if encode_base64:
            data = base64.b64encode(data)
            self.encoding = "base64"

        self.data = data

    def _decrypt(self):
        """
        Authenticate with Vault server & returns value of the key from secret

        :returns: secret in plain text
        """
        client = VaultClient(self.vault_params)

        # data is always base64 encoded
        data = base64.b64decode(self.data)

        # token will comprise of two parts, e.g. path/in/vault:key
        data_attrs = data.decode().split(":")
        if len(data_attrs) != 2:
            raise RefError(
                "Invalid vault secret: secret should be stored as 'path/in/vault:key', not '{}'".format(
                    data.decode()
                )
            )
        mount = self.vault_params.mount
        secret_path = data_attrs[0]
        secret_key = data_attrs[1]

        return_data = ""
        try:
            if self.vault_params.engine == VaultEngineTypes.KV:
                response = client.secrets.kv.v1.read_secret(
                    path=secret_path,
                    mount_point=mount,
                )
                return_data = response["data"][secret_key]
            else:
                response = client.secrets.kv.v2.read_secret_version(
                    path=secret_path,
                    mount_point=mount,
                    raise_on_deleted_version=False,
                )
                return_data = response["data"]["data"][secret_key]
            client.adapter.close()
        except Forbidden:
            raise VaultError(
                "Permission Denied. "
                + "make sure the token is authorised to access '{}' on Vault".format(secret_path)
            )
        except InvalidPath:
            raise VaultError("path '{}' does not exist on Vault".format(secret_path))
        except KeyError:
            raise VaultError("key '{}' does not exist on Vault".format(secret_key))
        finally:
            client.adapter.close()

        if return_data == "":
            raise VaultError("'{}' doesn't exist on '{}'".format(secret_key, secret_path))

        if self.encoding == "base64":
            return_data = base64.b64decode(return_data, validate=True).decode()

        return return_data

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {
            "data": self.data,
            "encoding": self.encoding,
            "type": self.type_name,
            "vault_params": self.vault_params.model_dump(),
        }

    def __str__(self):
        return "{}".format(self.dump())


class VaultBackend(Base64RefBackend):
    def __init__(self, path, ref_type=VaultSecret, **ref_kwargs):
        "init VaultBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = KapitanReferencesTypes.VAULTKV



from typing import List, Literal, Optional, Union

from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from pydantic_settings import BaseSettings

from kapitan.utils import StrEnum


class KapitanReferenceBaseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")


class KapitanReferenceGPGConfig(KapitanReferenceBaseConfig):
    recipients: List[dict[str, str]] = []


# Must be pydantic_settings.BaseSettings so that environment variables are actually used to
# populate the object. Any of the validation alias choices can be used as environment variable
# names.
# Note that this will break if both alias choices are set for a field (either through envvar or
# initializer).
class KapitanReferenceVaultEnv(BaseSettings):
    addr: Optional[str] = Field(None, validation_alias=AliasChoices("addr", "VAULT_ADDR"))
    skip_verify: Optional[bool] = Field(
        True, validation_alias=AliasChoices("skip_verify", "VAULT_SKIP_VERIFY")
    )
    client_key: Optional[str] = Field(None, validation_alias=AliasChoices("client_key", "VAULT_CLIENT_KEY"))
    client_cert: Optional[str] = Field(
        None, validation_alias=AliasChoices("client_cert", "VAULT_CLIENT_CERT")
    )
    cacert: Optional[str] = Field(None, validation_alias=AliasChoices("cacert", "VAULT_CACERT"))
    capath: Optional[str] = Field(None, validation_alias=AliasChoices("capath", "VAULT_CAPATH"))
    namespace: Optional[str] = Field(None, validation_alias=AliasChoices("namespace", "VAULT_NAMESPACE"))


class VaultEngineTypes(StrEnum):
    KV = "kv"
    KV_V2 = "kv-v2"
    TRANSIT = "transit"


class KapitanReferenceVaultCommon(KapitanReferenceVaultEnv):
    model_config = ConfigDict(use_enum_values=True)
    engine: Optional[VaultEngineTypes] = None
    auth: Optional[str] = None
    crypto_key: Optional[str] = None
    always_latest: Optional[bool] = False
    mount: Optional[str] = None
    key: Optional[str] = None


class KapitanReferenceVaultKVConfig(KapitanReferenceVaultCommon):
    engine: Literal[VaultEngineTypes.KV, VaultEngineTypes.KV_V2] = VaultEngineTypes.KV_V2
    mount: Optional[str] = "secret"


class KapitanReferenceVaultTransitConfig(KapitanReferenceVaultCommon):
    engine: Literal[VaultEngineTypes.TRANSIT] = VaultEngineTypes.TRANSIT
    key: Optional[str] = None
    mount: Optional[str] = "transit"


class KapitanReferenceAWKMSConfig(KapitanReferenceBaseConfig):
    key: str


class KapitanReferenceGKMSConfig(KapitanReferenceBaseConfig):
    key: str


class KapitanReferenceAZKMSConfig(KapitanReferenceBaseConfig):
    key: str


class KapitanReferenceConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    gpg: Optional[KapitanReferenceGPGConfig] = None
    awskms: Optional[KapitanReferenceAWKMSConfig] = None
    vaultkv: Optional[KapitanReferenceVaultKVConfig] = None
    gkms: Optional[KapitanReferenceGKMSConfig] = None
    vaulttransit: Optional[KapitanReferenceVaultTransitConfig] = None
    azkms: Optional[KapitanReferenceAZKMSConfig] = None



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import copy
import logging
import os
import sys
from typing import Any

import yaml
from omegaconf import Container, ListMergeMode, Node, OmegaConf

logger = logging.getLogger(__name__)


def key(_node_: Node):
    """resolver function, that returns the name of its key"""
    return _node_._key()


def parentkey(_parent_: Node):
    """resolver function, that returns the name of its parent key"""
    return _parent_._key()


def fullkey(_node_: Node):
    """resolver function, that returns the full name of its key"""
    return _node_._get_full_key("")


def access_key_with_dots(*key: str, _root_: Container):
    """resolver function, that accesses a key with dots in it"""
    value = _root_
    for part in key:
        value = value[part]

    return value


def escape_interpolation(content: str):
    """resolver function that escapes an interpolation for the next resolving step"""
    return f"${{{content}}}"


def merge(*args):
    """resolver function, that merges omegaconf objects"""
    return OmegaConf.merge(*args, list_merge_mode=ListMergeMode.EXTEND)


def to_dict(input_list: list):
    """resolver function, that converts an object to a dict"""
    if type(input_list) is not list or OmegaConf.is_list(input_list):
        # warning: input is not list
        return input_list

    if not all([type(item) is list or OmegaConf.is_dict(item) for item in input_list]):
        # warning: some items are not dict
        return input_list

    return {dict_key: item[dict_key] for item in input_list for dict_key in item}


def to_list(input_obj: Any):
    """resolver function, that converts an object to a list"""
    if type(input_obj) is dict or OmegaConf.is_dict(input_obj):
        return [{k: v} for k, v in input_obj.items()]

    return list(input_obj)


def to_yaml(key: str, _root_: Container):
    content = OmegaConf.select(_root_, key)
    return yaml.dump(OmegaConf.to_container(content, resolve=True))


def default(*args):
    output = ""
    for arg in args[:-1]:
        output += "${oc.select:" + str(arg) + ","

    output += str(args[-1])
    output += "}" * (len(args) - 1)
    return output


def relpath(absolute_path: str, _node_: Node):
    """
    resolver function, that translates an absolute yaml-path to its relative path
    """

    node_parts = []
    path_parts = absolute_path.split(".")
    relative_path = ""

    i = 0
    node = _node_
    while node._key() is not None:
        node_parts.append(node._key())
        node = node._get_parent()
        i += 1

    node_parts.reverse()

    for idx, (path_part, node_path) in enumerate(zip(path_parts, node_parts)):
        if not path_part == node_path:
            rel_prefix = "." * (i - idx) if idx != 0 else ""
            relative_path = rel_prefix + ".".join(path_parts[idx:])
            break

    if not relative_path:
        # warning: self reference
        return "SELF REFERENCE DETECTED"

    relative_interpolation = "${" + relative_path + "}"

    return relative_interpolation


def write_to_key(destination: str, origin: str, _root_):
    """
    resolver function to write any content to different place in the inventory
    NOTE: Behavior for lists is not well-defined
    """
    # fetch and resolve content
    try:
        content = OmegaConf.select(_root_, origin)
        if not content:
            # warning: origin could not be found / empty content
            return "NOT FOUND"

        # resolve relative interpolations
        try:
            # replace with OC.to_object(), when it supports escaped interpolations (wip)
            # reference: https://github.com/omry/omegaconf/pull/1113
            config = copy.deepcopy(content)
            OmegaConf.resolve(config, True)
        except Exception as e:
            # resolver error
            logger.warning(e)
            return "ERROR WHILE RESOLVING"

        # remove when issue above is resolved
        OmegaConf.set_readonly(config, False, recursive=True)

        # write resolved content back to _root_
        OmegaConf.update(_root_, destination, config, merge=True, force_add=True)
    except Exception as e:
        raise e
    return "DONE"


def from_file(file_path: str):
    if os.path.isfile(file_path):
        with open(file_path, "r") as f:
            return f.read()
    else:
        logger.error(f"from_file: file {file_path} does not exist")
    return "FILE NOT EXISTS"


def filename(_node_: Node):
    return _node_._get_flag("filename")


def parent_filename(_parent_: Node):
    return _parent_._get_flag("filename")


def path(_node_: Node):
    return _node_._get_flag("path")


def parent_path(_parent_: Node):
    return _parent_._get_flag("path")


def condition_if(condition: str, config: dict):
    if bool(condition):
        return config
    else:
        return {}


def condition_if_else(condition: str, config_if: dict, config_else: dict):
    if bool(condition):
        return config_if
    else:
        return config_else


def condition_not(condition: str):
    return not bool(condition)


def condition_and(*conditions: str):
    return all(conditions)


def condition_or(*conditions: str):
    return any(conditions)


def condition_equal(*configs):
    return all(config == configs[0] for config in configs)


def register_resolvers() -> None:
    """register pre-defined and user-defined resolvers"""
    replace = True

    # yaml key utility functions
    OmegaConf.register_new_resolver("key", key, replace=replace)
    OmegaConf.register_new_resolver("parentkey", parentkey, replace=replace)
    OmegaConf.register_new_resolver("fullkey", fullkey, replace=replace)
    OmegaConf.register_new_resolver("relpath", relpath, replace=replace)

    # yaml object utility functions
    OmegaConf.register_new_resolver("access", access_key_with_dots, replace=replace)
    OmegaConf.register_new_resolver("escape", escape_interpolation, replace=replace)
    OmegaConf.register_new_resolver("merge", merge, replace=replace)
    OmegaConf.register_new_resolver("dict", to_dict, replace=replace)
    OmegaConf.register_new_resolver("list", to_list, replace=replace)
    OmegaConf.register_new_resolver("yaml", to_yaml, replace=replace)
    OmegaConf.register_new_resolver("add", lambda x, y: x + y, replace=replace)
    OmegaConf.register_new_resolver("default", default, replace=replace)
    OmegaConf.register_new_resolver("write", write_to_key, replace=replace)
    OmegaConf.register_new_resolver("from_file", from_file, replace=replace)
    OmegaConf.register_new_resolver("filename", filename, replace=replace)
    OmegaConf.register_new_resolver("parent_filename", parent_filename, replace=replace)
    OmegaConf.register_new_resolver("path", path, replace=replace)
    OmegaConf.register_new_resolver("parent_path", parent_path, replace=replace)

    # boolean algebra
    OmegaConf.register_new_resolver("if", condition_if, replace=replace)
    OmegaConf.register_new_resolver("ifelse", condition_if_else, replace=replace)
    OmegaConf.register_new_resolver("and", condition_and, replace=replace)
    OmegaConf.register_new_resolver("or", condition_or, replace=replace)
    OmegaConf.register_new_resolver("not", condition_not, replace=replace)
    OmegaConf.register_new_resolver("equal", condition_equal, replace=replace)

    # user defined resolvers
    user_resolver_file = os.path.join(os.getcwd(), "system/omegaconf/resolvers/resolvers.py")

    if os.path.exists(user_resolver_file):
        try:
            register_user_resolvers(user_resolver_file)
        except Exception as e:
            logger.warning(f"Couldn't import {user_resolver_file}: error {e}")


def register_user_resolvers(user_resolver_file: str) -> None:
    """import user resolvers specified in resolvers_path"""
    if not os.path.exists(user_resolver_file):
        logger.debug(f"File {user_resolver_file} does not exist, ignoring")
        return

    import_path = os.path.dirname(user_resolver_file)

    try:
        sys.path.append(import_path)
        from resolvers import pass_resolvers

        funcs = pass_resolvers()
    except ImportError:
        logger.warning("resolvers.py must contain function 'pass_resolvers()'")
        return
    except Exception as e:
        logger.error(f"resolvers.py: {e}")
        return

    if not isinstance(funcs, dict):
        logger.warning("pass_resolvers() should return a dict")
        return

    import resolvers

    for name, func in funcs.items():
        try:
            OmegaConf.register_new_resolver(name, func, replace=True)
        except:
            logger.warning(f"Could not load resolver {name}")



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os

from kapitan import cached
from kapitan.inputs.base import CompiledFile, InputType
from kapitan.inventory.model.input_types import KapitanInputTypeJinja2Config
from kapitan.utils import render_jinja2

logger = logging.getLogger(__name__)


class Jinja2(InputType):

    def compile_file(self, config: KapitanInputTypeJinja2Config, input_path, compile_path):
        """
        Compile Jinja2 templates.

        Write items in ``input_path`` as Jinja2 rendered files to ``compile_path``.
        ``input_path`` can be either a file or directory.  The rendered files will be written
        to ``compile_path``.

        Args:
            config (KapitanInputTypeJinja2Config): Jinja2 input configuration.
            input_path (str): Path to the input Jinja2 template file or directory.
            compile_path (str): Path to write the compiled files.

        See ``kapitan.inputs.base.InputType`` for details on ``reveal`` and ``target_name``.
        """
        strip_postfix = config.suffix_remove
        stripped_postfix = config.suffix_stripped
        input_params = config.input_params

        reveal = self.args.reveal
        target_name = self.target_name

        # set compile_path allowing jsonnet to have context on where files
        # are being compiled during the current Kapitan run.  This is only done if the user
        # did not provide their own value.
        input_params.setdefault("compile_path", compile_path)

        # set ext_vars and inventory for jinja2 context
        context = {}

        context["inventory_global"] = cached.global_inv
        context["inventory"] = cached.global_inv[target_name]
        context["input_params"] = input_params
        vars = cached.global_inv[target_name]["parameters"]["kapitan"]["vars"]
        context.update(vars)

        jinja2_filters = self.args.jinja2_filters

        for item_key, item_value in render_jinja2(
            input_path, context, jinja2_filters=jinja2_filters, search_paths=self.search_paths
        ).items():
            if strip_postfix and item_key.endswith(stripped_postfix):
                item_key = item_key.rstrip(stripped_postfix)
            full_item_path = os.path.join(compile_path, item_key)
            with CompiledFile(
                full_item_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name
            ) as fp:
                fp.write(item_value["content"])
                mode = item_value["mode"]
                os.chmod(full_item_path, mode)
                logger.debug("Wrote %s with mode %.4o", full_item_path, mode)



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import hashlib
import logging
import multiprocessing
import os
from collections import defaultdict, namedtuple
from functools import partial
from mimetypes import MimeTypes
from shutil import copyfile, rmtree

from git import GitCommandError, Repo

from kapitan.errors import GitFetchingError, GitSubdirNotFoundError, HelmFetchingError
from kapitan.helm_cli import helm_cli
from kapitan.inventory.model.dependencies import KapitanDependencyTypes
from kapitan.utils import (
    copy_tree,
    make_request,
    normalise_join_path,
    safe_copy_file,
    safe_copy_tree,
    unpack_downloaded_file,
)

logger = logging.getLogger(__name__)

HelmSource = namedtuple("HelmSource", ["repo", "chart_name", "version", "helm_path"])


def fetch_dependencies(output_path, target_objs, save_dir, force, pool):
    """
    parses through the dependencies parameters in target_objs and fetches them
    all dependencies are first fetched into save_dir, after which they are copied to their respective output_path.
    overwites older version of existing dependencies if force fetced
    """
    # there could be multiple dependency items per source_uri due to reclass inheritance or
    # other user requirements. So create a mapping from source_uri to a set of dependencies with
    # that source_uri
    git_deps = defaultdict(list)
    http_deps = defaultdict(list)
    helm_deps = defaultdict(list)

    # this dict is to make sure no duplicated output_paths exist per source
    deps_output_paths = defaultdict(set)
    for target_obj in target_objs:
        try:
            dependencies = target_obj.dependencies
            for item in dependencies:
                dependency_type = item.type
                source_uri = item.source

                # The target key "output_path" is relative to the compile output path set by the user
                # point to the full output path
                full_output_path = normalise_join_path(output_path, item.output_path)
                logger.debug("Updated output_path from %s to %s", item.output_path, output_path)
                item.output_path = full_output_path

                if full_output_path in deps_output_paths[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    logger.debug("Skipping duplicate output path for uri %s", source_uri)
                    continue
                else:
                    deps_output_paths[source_uri].add(full_output_path)

                if dependency_type == KapitanDependencyTypes.GIT:
                    git_deps[source_uri].append(item)
                elif dependency_type in (KapitanDependencyTypes.HTTP, KapitanDependencyTypes.HTTPS):
                    http_deps[source_uri].append(item)
                elif dependency_type == KapitanDependencyTypes.HELM:
                    version = item.version
                    helm_deps[HelmSource(source_uri, item.chart_name, version, item.helm_path)].append(item)
                else:
                    logger.warning("%s is not a valid source type", dependency_type)

        except KeyError:
            logger.debug(
                "Target object %s has no 'dependencies' key, continuing", target_obj["vars"]["target"]
            )
            continue

    git_worker = partial(fetch_git_dependency, save_dir=save_dir, force=force)
    http_worker = partial(fetch_http_dependency, save_dir=save_dir, force=force)
    helm_worker = partial(fetch_helm_chart, save_dir=save_dir, force=force)
    [p.get() for p in pool.imap_unordered(http_worker, http_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(git_worker, git_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(helm_worker, helm_deps.items()) if p]


def fetch_git_dependency(dep_mapping, save_dir, force, item_type="Dependency"):
    """
    fetches a git repository at source into save_dir, and copy the repository into
    output_path stored in dep_mapping. ref is used to checkout if exists, fetches master branch by default.
    only subdir is copied into output_path if specified.
    """
    source, deps = dep_mapping
    # to avoid collisions between basename(source)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    cached_repo_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    if not exists_in_cache(cached_repo_path) or force:
        fetch_git_source(source, cached_repo_path, item_type)
    else:
        logger.debug("Using cached %s %s", item_type, cached_repo_path)
    for dep in deps:
        repo = Repo(cached_repo_path)
        output_path = dep.output_path
        copy_src_path = cached_repo_path
        repo.git.checkout(dep.ref)

        # initialising submodules
        if dep.submodules:
            for submodule in repo.submodules:
                submodule.update(init=True)

        if dep.subdir:
            sub_dir = dep.subdir
            full_subdir = os.path.join(cached_repo_path, sub_dir)
            if os.path.isdir(full_subdir):
                copy_src_path = full_subdir
            else:
                raise GitSubdirNotFoundError(
                    "{} {}: subdir {} not found in repo".format(item_type, source, sub_dir)
                )
        if force:
            # We need `clobber_files=True` here, since we otherwise can't overwrite read-only Git
            # index and pack files if the destination already contains a copy of the Git repo.
            copied = copy_tree(copy_src_path, output_path, clobber_files=True)
        else:
            copied = safe_copy_tree(copy_src_path, output_path)
        if copied:
            logger.info("%s %s: saved to %s", item_type, source, output_path)


def fetch_git_source(source, save_dir, item_type):
    """clones a git repository at source and saves into save_dir"""

    if os.path.exists(save_dir):
        rmtree(save_dir)
        logger.debug("Removed %s", save_dir)
    logger.debug("%s %s: fetching now", item_type, source)
    try:
        Repo.clone_from(source, save_dir)
        logger.debug("%s %s: successfully fetched", item_type, source)
        logger.debug("Git clone cached to %s", save_dir)
    except GitCommandError as e:
        logger.error(e)
        raise GitFetchingError("{} {}: fetching unsuccessful\n{}".format(item_type, source, e.stderr))


def fetch_http_dependency(dep_mapping, save_dir, force, item_type="Dependency"):
    """
    fetches a http[s] file at source and saves into save_dir, after which it is copied into
    the output_path stored in dep_mapping
    """
    source, deps = dep_mapping
    # to avoid collisions between basename(source)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    cached_source_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    if not exists_in_cache(cached_source_path) or force:
        content_type = fetch_http_source(source, cached_source_path, item_type)
    else:
        logger.debug("Using cached %s %s", item_type, cached_source_path)
        content_type = MimeTypes().guess_type(cached_source_path)[0]
    for dep in deps:
        output_path = dep.output_path
        if dep.unpack:
            # ensure that the directory we are extracting to exists
            os.makedirs(output_path, exist_ok=True)
            if force:
                is_unpacked = unpack_downloaded_file(cached_source_path, output_path, content_type)
            else:
                unpack_output = os.path.join(
                    save_dir, "extracted-" + str(multiprocessing.current_process().name)
                )
                os.makedirs(unpack_output)
                is_unpacked = unpack_downloaded_file(cached_source_path, unpack_output, content_type)
                safe_copy_tree(unpack_output, output_path)
                # delete unpack output
                rmtree(unpack_output)
            if is_unpacked:
                logger.info("%s %s: extracted to %s", item_type, source, output_path)
            else:
                logger.info(
                    "%s %s: Content-Type %s is not supported for unpack. Ignoring save",
                    item_type,
                    source,
                    content_type,
                )
        else:
            # we are downloading a single file
            parent_dir = os.path.dirname(output_path)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)
            if force:
                copyfile(cached_source_path, output_path)
            else:
                safe_copy_file(cached_source_path, output_path)
            logger.debug("%s %s: saved to %s", item_type, source, output_path)


def fetch_http_source(source, save_path, item_type):
    """downloads a http[s] file from source and saves into save_path(which is a file)"""

    if os.path.exists(save_path):
        os.remove(save_path)
        logger.debug("Removed %s", save_path)
    logger.debug("%s %s: fetching now", item_type, source)
    content, content_type = make_request(source)
    logger.debug("%s %s: successfully fetched", item_type, source)
    if content is not None:
        with open(save_path, "wb") as f:
            f.write(content)
        logger.debug("Cached to %s", save_path)
        return content_type
    logger.warning("%s %s: failed to fetch", item_type, source)
    return None


def fetch_helm_chart(dep_mapping, save_dir, force):
    """
    downloads a helm chart and its subcharts from source then untars and moves it to save_dir
    """
    source, deps = dep_mapping

    # to avoid collisions between source.chart_name/source.version
    path_hash = hashlib.sha256(source.repo.encode()).hexdigest()[:8]
    cached_repo_path = os.path.join(
        save_dir, path_hash, source.chart_name + "-" + (source.version or "latest")
    )

    for dep in deps:
        output_path = dep.output_path

        if not os.path.exists(output_path) or force:
            if not exists_in_cache(cached_repo_path):
                fetch_helm_archive(
                    source.helm_path, source.repo, source.chart_name, source.version, cached_repo_path
                )
            else:
                logger.debug("Using cached helm chart at %s", cached_repo_path)

            parent_dir = os.path.dirname(output_path)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)

            if force:
                copied = copy_tree(cached_repo_path, output_path)
            else:
                copied = safe_copy_tree(cached_repo_path, output_path)

            if copied:
                logger.info("Dependency %s: saved to %s", source.chart_name, output_path)


def fetch_helm_archive(helm_path, repo, chart_name, version, save_path):
    logger.debug("Dependency helm chart %s and version %s: fetching now", chart_name, version or "latest")
    # Fetch archive and untar it into parent dir
    save_dir = os.path.dirname(save_path)
    args = ["pull", "--destination", save_dir, "--untar"]

    if version:
        args.append("--version")
        args.append(version)

    if repo.startswith("oci://"):
        args.append(repo)
    else:
        args.append("--repo")
        args.append(repo)
        args.append(chart_name)

    response = helm_cli(helm_path, args)
    if response != "":
        logger.warning("Dependency helm chart %s and version %s: %s", chart_name, version, response)
        raise HelmFetchingError(response)
    else:
        # rename chart to requested name
        os.rename(os.path.join(save_dir, chart_name), save_path)
        logger.debug("Dependency helm chart %s and version %s: successfully fetched", chart_name, version)
        logger.debug("Dependency helm chart %s and version %s: saved to %s", chart_name, version, save_path)


def exists_in_cache(item_path):
    dep_cache_path = os.path.dirname(item_path)
    if not os.path.exists(dep_cache_path):
        return False
    return os.path.basename(item_path) in os.listdir(dep_cache_path)



from typing import Type

from kapitan.utils import StrEnum

from .inventory import Inventory, InventoryError, InventoryTarget


class InventoryBackends(StrEnum):
    """
    Enumeration of available inventory backends.
    """

    RECLASS = "reclass"
    RECLASS_RS = "reclass-rs"
    OMEGACONF = "omegaconf"
    DEFAULT = RECLASS


def load_reclass_backend():
    """
    Enable the reclass inventory backend.
    """
    from .backends.reclass import ReclassInventory

    return ReclassInventory


def load_reclass_rs_backend():
    """
    Enable the reclass-rs inventory backend.
    """
    from .backends.reclass_rs import ReclassRsInventory

    return ReclassRsInventory


def load_omegaconf_backend():
    """
    Enable the omegaconf inventory backend.
    """
    from .backends.omegaconf import OmegaConfInventory

    return OmegaConfInventory


# Dict mapping values for command line flag `--inventory-backend` to the
# associated `Inventory` subclass.
AVAILABLE_BACKENDS: dict[str, Type[Inventory]] = {
    InventoryBackends.RECLASS: load_reclass_backend,
    InventoryBackends.RECLASS_RS: load_reclass_rs_backend,
    InventoryBackends.OMEGACONF: load_omegaconf_backend,
}


def get_inventory_backend(backend_name: str) -> Type[Inventory]:
    """
    Get the `Inventory` subclass associated with the given `backend_name`.
    """
    return AVAILABLE_BACKENDS.get(backend_name, AVAILABLE_BACKENDS[InventoryBackends.DEFAULT])()



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"hashicorp vault resource functions"

import logging
import os

import hvac

from kapitan.errors import KapitanError
from kapitan.inventory.model.references import KapitanReferenceVaultCommon

logger = logging.getLogger(__name__)


class VaultError(KapitanError):
    """Generic vault errors"""


class VaultClient(hvac.Client):
    """client connects to vault server and authenticates itself"""

    def __init__(self, vault_params: KapitanReferenceVaultCommon):
        if isinstance(vault_params, dict):
            vault_params = KapitanReferenceVaultCommon(**vault_params)
        self.vault_params = vault_params
        self.env = get_env(vault_params)

        super().__init__(**{k: v for k, v in self.env.items() if k != "auth"})

        self.authenticate()

    def get_auth_token(self):
        """
        get token either from environment or from file
        """
        auth_type = self.vault_params.auth
        token = ""

        if auth_type in ["token", "github"]:
            token = os.getenv("VAULT_TOKEN")
            if not token:
                token_file = os.path.join(os.path.expanduser("~"), ".vault-token")
                token = self.read_token_from_file(token_file)

        self.env["token"] = token

    def read_token_from_file(self, token_file):
        try:
            with open(token_file, "r") as fp:
                token = fp.read()
        except IOError:
            raise VaultError("Cannot read file {}".format(token_file))

        if not token:
            raise VaultError("{} is empty".format(token_file))

        return token

    def authenticate(self):
        # DIFFERENT LOGIN METHOD BASED ON AUTHENTICATION TYPE
        auth_type = self.vault_params.auth
        self.get_auth_token()
        username = os.getenv("VAULT_USERNAME")
        password = os.getenv("VAULT_PASSWORD")

        if auth_type == "token":
            self.token = self.env["token"]
        elif auth_type == "ldap":
            self.auth.ldap.login(username=username, password=password)
        elif auth_type == "userpass":
            self.auth.userpass.login(username=username, password=password)
        elif auth_type == "approle":
            self.auth.approle.login(os.getenv("VAULT_ROLE_ID"), secret_id=os.getenv("VAULT_SECRET_ID"))
        elif auth_type == "github":
            self.auth.github.login(token=self.env["token"])
        else:
            raise VaultError("Authentication type '{}' not supported".format(auth_type))

        if not self.is_authenticated():
            self.adapter.close()
            raise VaultError(
                "Vault Authentication Error, check if token in env:VAULT_TOKEN is valid and not expired"
            )


def get_env(vault_params: KapitanReferenceVaultCommon):
    """
    The following variables need to be exported to the environment or defined in inventory.
        * VAULT_ADDR: url for vault
        * VAULT_SKIP_VERIFY=true: if set, do not verify presented TLS certificate before communicating with Vault server.
        * VAULT_CLIENT_KEY: path to an unencrypted PEM-encoded private key matching the client certificate
        * VAULT_CLIENT_CERT: path to a PEM-encoded client certificate for TLS authentication to the Vault server
        * VAULT_CACERT: path to a PEM-encoded CA cert file to use to verify the Vault server TLS certificate
        * VAULT_CAPATH: path to a directory of PEM-encoded CA cert files to verify the Vault server TLS certificate
        * VAULT_NAMESPACE: specify the Vault Namespace, if you have one
    Following keys are used to creates a new hvac client instance.
        :param url: Base URL for the Vault instance being addressed.
        :type url: str
        :param cert: Certificates for use in requests sent to the Vault instance. This should be a tuple with the
            certificate and then key.
        :type cert: tuple
        :param verify: Either a boolean to indicate whether TLS verification should be performed when sending requests to Vault,
            or a string pointing at the CA bundle to use for verification.
            ### ssl-cert-verification.
            See http://docs.python-requests.org/en/master/user/advanced/
        :type verify: Union[bool,str]
        :param namespace: Optional Vault Namespace.
        :type namespace: str
    """
    if isinstance(vault_params, dict):
        vault_params = KapitanReferenceVaultCommon(**vault_params)

    client_parameters = {}
    client_parameters["url"] = vault_params.addr
    client_parameters["namespace"] = vault_params.namespace

    # VERIFY VAULT SERVER TLS CERTIFICATE
    skip_verify = vault_params.skip_verify

    if not skip_verify:
        cert = vault_params.cacert
        if not cert:
            cert_path = vault_params.capath
            if not cert_path:
                raise VaultError("Neither VAULT_CACERT nor VAULT_CAPATH specified")
            client_parameters["verify"] = cert_path
        else:
            client_parameters["verify"] = cert
    else:
        client_parameters["verify"] = False

    # CLIENT CERTIFICATE FOR TLS AUTHENTICATION
    client_key = vault_params.client_key
    client_cert = vault_params.client_cert
    if client_key != "" and client_cert != "":
        client_parameters["cert"] = (client_cert, client_key)
    return client_parameters



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os

from kapitan.errors import CompileError
from kapitan.inputs.base import InputType
from kapitan.inventory.model.input_types import KapitanInputTypeJsonnetConfig
from kapitan.resources import resource_callbacks, search_imports

logger = logging.getLogger(__name__)


def select_jsonnet_runtime(use_go):
    """Selects the Jsonnet runtime to use.

    Args:
        use_go: Boolean indicating whether to use the Go Jsonnet runtime.

    Returns:
        module: The selected Jsonnet runtime module (_jsonnet or _gojsonnet).

    Raises:
        ImportError: If the selected Jsonnet runtime is not installed.  This will happen if neither
                     the python or go jsonnet libraries are installed.  The error message will
                     provide instructions on how to install the necessary library.
    Raises:
        ImportError: If the _jsonnet module is not installed.
        CompileError: If there is an error during Jsonnet evaluation.
    """
    try:
        if use_go:
            import _gojsonnet

            return _gojsonnet
        else:
            import _jsonnet

            return _jsonnet

    except ImportError as exc:
        raise ImportError(
            "Jsonnet is not installed or running on an unsupported architecture. "
            "See https://kapitan.dev/compile/#jsonnet for installation instructions."
        ) from exc


class Jsonnet(InputType):
    """Jsonnet input type."""

    def compile_file(self, config: KapitanInputTypeJsonnetConfig, input_path: str, compile_path: str):
        """Compiles a Jsonnet file and writes the output to the compile directory.

        Args:
            config: KapitanInputTypeJsonnetConfig object containing compilation options.
            input_path: Path to the Jsonnet file to compile.
            compile_path: Path to the directory where compiled output will be written.

        Raises:
            CompileError: If there is an error compiling or writing the output.
            ValueError: If the specified output type is invalid.

        This function evaluates the Jsonnet file, handles different output formats (json, yaml, toml, plain),
        prunes empty values if requested, and writes the results to individual files in the
        compile directory.
        """

        use_go = self.args.use_go_jsonnet
        ext_vars = {"target": self.target_name}

        def _search_imports(cwd, imp):
            return search_imports(cwd, imp, self.search_paths)

        try:
            jsonnet = select_jsonnet_runtime(use_go)
            json_output = jsonnet.evaluate_file(
                input_path,
                import_callback=_search_imports,
                native_callbacks=resource_callbacks(self.search_paths),
                ext_vars=ext_vars,
            )

            output_obj = json.loads(json_output)

        except (ImportError, CompileError) as e:
            raise CompileError(f"Jsonnet Error compiling {input_path}: {e}") from e

        # If output_obj is not a dictionary, wrap it in a dictionary using the input filename
        # (without extension) as the key. This ensures that even single-item outputs are handled correctly.
        if not isinstance(output_obj, dict):
            filename = os.path.splitext(os.path.basename(input_path))[0]
            # Using filename as key ensures that single-item outputs are handled correctly.
            # Prevents issues when a single item is returned and needs to be written to a file.
            output_obj = {filename: output_obj}

        # Write each item in output_obj to a separate file.
        for item_key, item_value in output_obj.items():
            file_path = os.path.join(compile_path, item_key)
            self.to_file(config, file_path, item_value)



import logging
from typing import Annotated, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from kapitan.utils import StrEnum

logger = logging.getLogger(__name__)


class InputTypes(StrEnum):
    JSONNET = "jsonnet"
    JINJA2 = "jinja2"
    HELM = "helm"
    KADET = "kadet"
    COPY = "copy"
    REMOVE = "remove"
    EXTERNAL = "external"


class OutputType(StrEnum):
    JSON = "json"
    YAML = "yaml"
    YML = "yml"
    PLAIN = "plain"
    TOML = "toml"
    AUTO = "auto"


class KapitanInputTypeBaseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    name: Optional[str] = Field(default=None)
    input_type: InputTypes
    input_paths: List[str]
    output_path: str
    input_params: dict = {}
    continue_on_compile_error: Optional[bool] = False

    output_type: OutputType = OutputType.YAML
    ignore_missing: Optional[bool] = False
    prune: Optional[bool] = True


class KapitanInputTypeJsonnetConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.JSONNET] = InputTypes.JSONNET
    prune: Optional[bool] = False


class KapitanInputTypeExternalConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.EXTERNAL] = InputTypes.EXTERNAL
    env_vars: dict[str, str] = {}
    args: List[str] = []


class KapitanInputTypeCopyConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.COPY] = InputTypes.COPY
    ignore_missing: bool = False


class KapitanInputTypeJinja2Config(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.JINJA2] = InputTypes.JINJA2
    output_type: Optional[OutputType] = OutputType.PLAIN
    ignore_missing: Optional[bool] = True
    suffix_remove: Optional[bool] = False
    suffix_stripped: Optional[str] = ".j2"


class KapitanInputTypeHelmConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.HELM] = InputTypes.HELM
    output_type: OutputType = OutputType.AUTO
    prune: Optional[bool] = False
    helm_params: dict = {}
    helm_values: Optional[dict] = {}
    helm_values_files: Optional[List[str]] = []
    helm_path: Optional[str] = None
    kube_version: Optional[str] = None

    @field_validator("output_type")
    @classmethod
    def type_must_be_auto(cls, _: OutputType) -> OutputType:
        """Helm output type must be auto."""
        logger.debug("field `output_type` for helm input type must be set to 'auto': enforcing.")
        return OutputType.AUTO


class KapitanInputTypeKadetConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.KADET] = InputTypes.KADET
    output_type: OutputType = OutputType.YAML
    input_value: Optional[dict] = None
    prune: Optional[bool] = False


class KapitanInputTypeRemoveConfig(KapitanInputTypeBaseConfig):
    input_type: Literal[InputTypes.REMOVE] = InputTypes.REMOVE
    output_path: Optional[str] = None


CompileInputTypeConfig = Annotated[
    Union[
        KapitanInputTypeJinja2Config,
        KapitanInputTypeExternalConfig,
        KapitanInputTypeCopyConfig,
        KapitanInputTypeKadetConfig,
        KapitanInputTypeJsonnetConfig,
        KapitanInputTypeHelmConfig,
        KapitanInputTypeRemoveConfig,
    ],
    Field(discriminator="input_type"),
]



from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field

from kapitan.inventory.model.dependencies import DependencyTypeConfig
from kapitan.inventory.model.input_types import CompileInputTypeConfig
from kapitan.inventory.model.references import KapitanReferenceConfig


class KapitanEssentialVars(BaseModel):
    model_config = ConfigDict(extra="allow")
    target: Optional[str] = None


class KapitanInventorySettings(BaseModel):
    compile: List[CompileInputTypeConfig] = []
    vars: KapitanEssentialVars = KapitanEssentialVars()
    labels: dict[str, str] = {}
    dependencies: Optional[List[DependencyTypeConfig]] = []
    target_full_path: str = ""
    secrets: Optional[KapitanReferenceConfig] = None
    validate_: list[dict] = Field(alias="validate", default=[])


class KapitanMetadataName(BaseModel):
    short: str
    full: str
    path: str
    parts: List[str]


class KapitanInventoryMetadata(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)
    name: Optional[KapitanMetadataName] = None


class KapitanInventoryParameters(BaseModel):
    model_config = ConfigDict(extra="allow")

    kapitan: Optional[KapitanInventorySettings] = KapitanInventorySettings()
    kapitan_metadata: Optional[KapitanInventoryMetadata] = Field(alias="_kapitan_", default=None)
    reclass_metadata: Optional[KapitanInventoryMetadata] = Field(alias="_reclass_", default=None)



import logging
import os
from datetime import datetime

import reclass
import reclass.core
import yaml
from reclass.errors import NotFoundError, ReclassException

from kapitan.errors import InventoryError
from kapitan.inventory import Inventory, InventoryTarget

logger = logging.getLogger(__name__)


class ReclassInventory(Inventory):

    def render_targets(
        self, targets: list[InventoryTarget] = None, ignore_class_not_found: bool = False
    ) -> None:
        """
        Runs a reclass inventory in inventory_path
        (same output as running ./reclass.py -b inv_base_uri/ --inventory)
        Will attempt to read reclass config from 'reclass-config.yml' otherwise
        it will fall back to the default config.
        Returns a reclass style dictionary

        Does not throw errors if a class is not found while ignore_class_not_found is specified
        """
        reclass_config = get_reclass_config(
            self.inventory_path, ignore_class_not_found, self.compose_target_name
        )

        try:
            storage = reclass.get_storage(
                reclass_config["storage_type"],
                reclass_config["nodes_uri"],
                reclass_config["classes_uri"],
                reclass_config["compose_node_name"],
            )
            class_mappings = reclass_config.get("class_mappings")  # this defaults to None (disabled)
            _reclass = reclass.core.Core(storage, class_mappings, reclass.settings.Settings(reclass_config))
            start = datetime.now()
            rendered_inventory = _reclass.inventory()
            elapsed = datetime.now() - start
            logger.debug(f"Inventory rendering with reclass took {elapsed}")

            # store parameters and classes
            for target_name, rendered_target in rendered_inventory["nodes"].items():
                self.targets[target_name].parameters = rendered_target["parameters"]
                self.targets[target_name].classes = rendered_target["classes"]
                self.targets[target_name].applications = rendered_target["applications"]
                self.targets[target_name].exports = rendered_target["exports"]

        except ReclassException as e:
            if isinstance(e, NotFoundError):
                logger.error("Inventory reclass error: inventory not found")
            else:
                logger.error(f"Inventory reclass error: {e.message}")
            raise InventoryError(e.message)


def get_reclass_config(
    inventory_path: str,
    ignore_class_not_found: bool = False,
    compose_target_name: bool = False,
    normalise_nodes_classes: bool = True,
) -> dict:
    # set default values initially
    reclass_config = {
        "storage_type": "yaml_fs",
        "inventory_base_uri": inventory_path,
        "nodes_uri": "targets",
        "classes_uri": "classes",
        "compose_node_name": compose_target_name,
        "allow_none_override": True,
        "ignore_class_notfound": ignore_class_not_found,  # reclass has it mispelled
    }
    try:
        from yaml import CSafeLoader as YamlLoader
    except ImportError:
        from yaml import SafeLoader as YamlLoader

    # get reclass config from file 'inventory/reclass-config.yml'
    cfg_file = os.path.join(inventory_path, "reclass-config.yml")
    if os.path.isfile(cfg_file):
        with open(cfg_file, "r") as fp:
            config = yaml.load(fp.read(), Loader=YamlLoader)
            logger.debug(f"Using reclass inventory config at: {cfg_file}")
        if config:
            # set attributes, take default values if not present
            for key, value in config.items():
                reclass_config[key] = value
        else:
            logger.debug(
                f"Reclass config: Empty config file at {cfg_file}. Using reclass inventory config defaults"
            )
    else:
        logger.debug("Inventory reclass: No config file found. Using reclass inventory config defaults")

    # normalise relative nodes_uri and classes_uri paths
    if normalise_nodes_classes:
        for uri in ("nodes_uri", "classes_uri"):
            reclass_config[uri] = os.path.normpath(os.path.join(inventory_path, reclass_config[uri]))

    return reclass_config



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

from kapitan.utils import StrEnum


class KapitanReferencesTypes(StrEnum):
    GPG = "gpg"
    VAULTKV = "vaultkv"
    VAULTTRANSIT = "vaulttransit"
    AWSKMS = "awskms"
    GKMS = "gkms"
    AZKMS = "azkms"
    BASE64 = "base64"
    PLAIN = "plain"
    ENV = "env"



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import multiprocessing as mp
import os
import time
from functools import singledispatch

import yaml
from cachetools import LRUCache, cached
from kadet import Dict
from omegaconf import ListMergeMode, OmegaConf

from kapitan.errors import InventoryError
from kapitan.inventory import Inventory, InventoryTarget
from kapitan.inventory.model import KapitanInventoryMetadata, KapitanInventoryParameters

from .migrate import migrate
from .resolvers import register_resolvers

logger = logging.getLogger(__name__)


@singledispatch
def keys_to_strings(ob):
    return ob


class OmegaConfRenderingError(InventoryError):
    pass


@keys_to_strings.register
def _handle_dict(ob: dict):
    return {str(k): keys_to_strings(v) for k, v in ob.items()}


@keys_to_strings.register
def _handle_list(ob: list):
    return [keys_to_strings(v) for v in ob]


class OmegaConfTarget(InventoryTarget):
    resolved: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__add_metadata()

    def __add_metadata(self):
        name_metadata = {
            "short": self.name.split(".")[-1],
            "full": self.name,
            "path": os.path.splitext(self.path)[0],
            "parts": self.name.split("."),
        }
        kapitan_metadata = KapitanInventoryMetadata(name=name_metadata)

        self.parameters = KapitanInventoryParameters(_kapitan_=kapitan_metadata, _reclass_=kapitan_metadata)


class OmegaConfInventory(Inventory):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, target_class=OmegaConfTarget)

    def render_targets(
        self, targets: list[OmegaConfTarget] = None, ignore_class_not_found: bool = False
    ) -> None:
        if not self.initialised:
            manager = mp.Manager()
            shared_targets = manager.dict()
            with mp.Pool(min(len(targets), os.cpu_count())) as pool:
                r = pool.map_async(
                    self.inventory_worker, [(self, target, shared_targets) for target in targets.values()]
                )
                r.wait()

                if not r.successful():
                    raise OmegaConfRenderingError("Error while loading the OmegaConf inventory")

            for target in shared_targets.values():
                self.targets[target.name] = target

    @staticmethod
    def inventory_worker(zipped_args):
        self, target, shared_targets = zipped_args
        try:
            register_resolvers()
            self.load_target(target)
            shared_targets[target.name] = target

        except Exception as e:
            logger.error(f"{target.name}: could not render due to error {e}")
            raise

    @cached(cache=LRUCache(maxsize=1024))
    def resolve_class_file_path(
        self, class_name: str, class_parent_dir: str = None, class_parent_name: str = None
    ):
        class_file = None

        # Finds relative paths based on the parent directory
        if class_name.startswith(".") and class_parent_dir:
            class_path_base = os.path.join(self.classes_path, class_parent_dir)
        else:
            class_path_base = self.classes_path

        # Now try to find the class file
        extension = ".yml"

        cases = [
            # case components.kapicorp is absolute and a directory, look for components/kapicorp/init.yml
            # case .components.kapicorp is relative and a directory, look for <class_parent_dir>/components/kapicorp/init.yml
            os.path.join(class_path_base, *class_name.split("."), "init" + extension),
            # case components.kapicorp is absolute and a file, look for components/kapicorp.yml
            # case components.kapicorp is relative and a file, look for <class_parent_dir>/components/kapicorp.yml
            os.path.join(class_path_base, *class_name.split(".")) + extension,
            # Reclass compatibility mode
            # case .components.kapicorp  points to <class_parent_dir>/kapicorp.yml
            os.path.join(class_path_base, *class_name.split(".")[2:]) + extension,
            # case components.kapicorp points to components/kapicorp/init.yml
            os.path.join(class_path_base, *class_name.split(".")[2:], "init" + extension),
        ]

        for case in cases:
            if os.path.isfile(case):
                class_file = case
                return class_file

        logger.error(f"class file not found for class {class_name}, tried {cases}")
        return None

    @cached(cache=LRUCache(maxsize=1024))
    def load_file(self, filename):
        with open(filename, "r") as f:
            return yaml.safe_load(f)

    def load_parameters_from_file(self, filename, parameters={}) -> Dict:
        parameters = OmegaConf.create(parameters)
        applications = []
        classes = []
        exports = Dict()

        content = self.load_file(filename)

        _classes = content.get("classes", [])
        _parameters = keys_to_strings(content.get("parameters", {}))
        _applications = content.get("applications", [])
        _exports = content.get("exports", {})

        # first processes all classes
        for class_name in _classes:

            class_parent_dir = os.path.dirname(filename.removeprefix(self.classes_path).removeprefix("/"))
            class_parent_name = os.path.basename(filename)
            class_file = self.resolve_class_file_path(
                class_name, class_parent_dir=class_parent_dir, class_parent_name=class_parent_name
            )
            if not class_file:
                if self.ignore_class_not_found:
                    continue
                raise InventoryError(f"Class {class_name} not found")
            p, c, a, e = self.load_parameters_from_file(class_file)
            if p:
                parameters = OmegaConf.unsafe_merge(
                    parameters, p, list_merge_mode=ListMergeMode.EXTEND_UNIQUE
                )
            classes.extend(c)
            classes.append(class_name)
            applications.extend(a)
            exports.merge_update(e, box_merge_lists="unique")

        # finally merges the parameters from the current file
        if _parameters:
            parameters = OmegaConf.unsafe_merge(
                parameters, _parameters, list_merge_mode=ListMergeMode.EXTEND_UNIQUE
            )

        exports.merge_update(_exports, box_merge_lists="unique")
        applications.extend(_applications)
        return parameters, classes, applications, exports

    def load_target(self, target: OmegaConfTarget):
        full_target_path = os.path.join(self.targets_path, target.path)

        start = time.perf_counter()
        parameters = OmegaConf.create(keys_to_strings(target.parameters.model_dump(by_alias=True)))
        p, c, a, e = self.load_parameters_from_file(full_target_path, parameters=parameters)
        load_parameters = time.perf_counter() - start
        target.parameters = OmegaConf.to_container(p, resolve=True)
        to_container = time.perf_counter() - load_parameters
        target.classes = c
        target.applications = a
        target.exports = e
        finish_loading = time.perf_counter() - start
        logger.debug(
            f"{target.name}: Config loaded in {load_parameters:.2f}s, resolved in {to_container:.2f}s. Total {finish_loading:.2f}s"
        )

    def migrate(self):
        migrate(self.inventory_path)

    def resolve_targets(self, targets: list[OmegaConfTarget] = None) -> None:
        if not targets:
            targets = self.targets.values()
        map(lambda target: target.resolve(), targets)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import shutil

from kapitan.inputs.base import InputType
from kapitan.inventory.model.input_types import KapitanInputTypeCopyConfig
from kapitan.utils import copy_tree

logger = logging.getLogger(__name__)


class Copy(InputType):
    def compile_file(self, config: KapitanInputTypeCopyConfig, input_path, compile_path):
        """Copy input_path to compile_path.

        Args:
            config (KapitanInputTypeCopyConfig): input configuration.
            input_path (str): path to the file or directory to copy.
            compile_path (str): path to the destination directory.

        Raises:
            OSError: if input_path does not exist and ignore_missing is False.
        """

        ignore_missing = config.ignore_missing

        try:
            if os.path.exists(input_path):
                logger.debug("Copying '%s' to '%s'.", input_path, compile_path)
                if os.path.isfile(input_path):
                    if os.path.isfile(compile_path):
                        # overwrite existing file
                        shutil.copy2(input_path, compile_path)
                    else:
                        # create destination directory if it doesn't exist
                        os.makedirs(compile_path, exist_ok=True)
                        # copy file to destination directory
                        shutil.copy2(input_path, os.path.join(compile_path, os.path.basename(input_path)))
                else:
                    # Resolve relative paths to avoid issues with copy_tree
                    compile_path = os.path.abspath(compile_path)  # Resolve relative paths
                    copy_tree(input_path, compile_path)
            elif not ignore_missing:
                # Raise exception if input path does not exist and ignore_missing is False
                raise OSError(f"Path {input_path} does not exist and `ignore_missing` is {ignore_missing}")
        except OSError as e:
            # Log exception and re-raise
            logger.exception("Input dir not copied. Error: %s", e)



#!/usr/bin/env python3

# Copyright 2023 The Kapitan Authors
# SPDX-FileCopyrightText: 2023 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import functools
import logging
import os
from abc import ABC, abstractmethod
from typing import Dict

from pydantic import BaseModel, ConfigDict, Field

from kapitan.errors import InventoryError
from kapitan.inventory.model import KapitanInventoryParameters

logger = logging.getLogger(__name__)


class InventoryTarget(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True)
    name: str = Field(exclude=True)
    path: str = Field(exclude=True)
    parameters: KapitanInventoryParameters = KapitanInventoryParameters()
    classes: list = list()
    applications: list = list()
    exports: dict = {}


class Inventory(ABC):
    def __init__(
        self,
        inventory_path: str = "inventory",
        compose_target_name: bool = False,
        ignore_class_not_found=False,
        initialise=True,
        target_class=InventoryTarget,
    ):
        self.inventory_path = inventory_path
        self.compose_target_name = compose_target_name
        self.targets_path = os.path.join(self.inventory_path, "targets")
        self.classes_path = os.path.join(self.inventory_path, "classes")
        self.initialised: bool = False
        self.targets: dict[str, target_class] = {}
        self.ignore_class_not_found = ignore_class_not_found
        self.target_class = target_class

        if initialise:
            self.__initialise(ignore_class_not_found=ignore_class_not_found)

    @functools.cached_property
    def inventory(self) -> dict:
        """
        get all targets from inventory
        """

        return {target.name: target.model_dump(by_alias=True) for target in self.targets.values()}

    def __initialise(self, ignore_class_not_found) -> bool:
        """
        look for targets at '<inventory_path>/targets/' and initialise them.
        """
        logger.debug(f"Initialising inventory from {self.targets_path}")
        if not self.initialised:
            for root, _, files in os.walk(self.targets_path):
                for file in files:
                    # split file extension and check if yml/yaml
                    path = os.path.relpath(os.path.join(root, file), self.targets_path)

                    if self.compose_target_name:
                        name, ext = os.path.splitext(path)
                        name = name.replace(os.sep, ".")
                    else:
                        name, ext = os.path.splitext(file)

                    if ext not in (".yml", ".yaml"):
                        logger.debug(f"ignoring {file}: targets have to be .yml or .yaml files.")
                        continue

                    target = self.target_class(name=name, path=path)

                    if self.targets.get(target.name):
                        raise InventoryError(
                            f"Conflicting targets {target.name}: {target.path} and {self.targets[target.name].path}. "
                            f"Consider using '--compose-target-name'."
                        )

                    self.targets[target.name] = target

            self.render_targets(self.targets, ignore_class_not_found=ignore_class_not_found)
            self.initialised = True
        return self.initialised

    def get_target(self, target_name: str, ignore_class_not_found: bool = False) -> InventoryTarget:
        """
        helper function to get rendered InventoryTarget object for single target
        """
        return self.targets.get(target_name)

    def get_targets(self, target_names: list[str] = [], ignore_class_not_found: bool = False) -> dict:
        """
        helper function to get rendered InventoryTarget objects for multiple targets
        """

        if target_names:
            return {
                target_name: self.targets[target_name]
                for target_name in target_names
                if target_name in self.targets
            }
        else:
            return self.targets

    def get_parameters(self, target_names: str | list[str], ignore_class_not_found: bool = False) -> dict:
        """
        helper function to get rendered parameters for single target or multiple targets
        """
        if type(target_names) is str:
            target = self.get_target(target_names, ignore_class_not_found)
            return target.parameters

        return {
            name: {"parameters": Dict(target.parameters)} for name, target in self.get_targets(target_names)
        }

    @abstractmethod
    def render_targets(
        self, targets: list[InventoryTarget] = None, ignore_class_not_found: bool = False
    ) -> None:
        """
        create the inventory depending on which backend gets used
        """
        raise NotImplementedError

    def migrate(self):
        """
        migrate the inventory, e.g. change interpolation syntax to new syntax
        """

    def __getitem__(self, key):
        return self.inventory[key]



import logging
from datetime import datetime

import reclass_rs

from kapitan.errors import InventoryError
from kapitan.inventory import Inventory
from kapitan.inventory.backends.reclass import get_reclass_config

logger = logging.getLogger(__name__)


class ReclassRsInventory(Inventory):
    def _make_reclass_rs(self, ignore_class_not_found: bool):
        # Get Reclass config options with the same method that's used for `ReclassInventory`, but
        # disable the logic to normalise the `nodes_uri` and `classes_uri` options, since reclass-rs
        # expects those fields to be relative to the inventory path.
        config_dict = get_reclass_config(
            self.inventory_path, ignore_class_not_found, self.compose_target_name, False
        )

        # Turn on verbose config loading only if Kapitan loglevel is set to at least DEBUG.
        config = reclass_rs.Config.from_dict(
            self.inventory_path, config_dict, logger.isEnabledFor(logging.DEBUG)
        )
        return reclass_rs.Reclass.from_config(config)

    def render_targets(self, targets: list = None, ignore_class_not_found: bool = False):
        try:
            r = self._make_reclass_rs(ignore_class_not_found)
            start = datetime.now()
            inv = r.inventory()
            elapsed = datetime.now() - start
            logger.debug(f"Inventory rendering with reclass-rs took {elapsed}")

            for target_name, nodeinfo in inv.nodes.items():
                self.targets[target_name].parameters = nodeinfo.parameters
                self.targets[target_name].classes = nodeinfo.classes
                self.targets[target_name].applications = nodeinfo.applications
                self.targets[target_name].exports = nodeinfo.exports

        except ValueError as e:
            logger.error(f"Reclass-rs error: {e}")
            raise InventoryError(f"{e}")



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import functools  # Import functools for caching
import importlib
from typing import Type

from kapitan.inventory.model.input_types import InputTypes

from .base import InputType


@functools.lru_cache(maxsize=None)  # Use lru_cache for caching
def get_compiler(input_type: InputType) -> Type[InputType]:
    """Dynamically imports and returns the compiler class based on input_type."""

    module_map = {
        InputTypes.JINJA2: "jinja2",
        InputTypes.HELM: "helm",
        InputTypes.JSONNET: "jsonnet",
        InputTypes.KADET: "kadet",
        InputTypes.COPY: "copy",
        InputTypes.EXTERNAL: "external",
        InputTypes.REMOVE: "remove",
    }

    module_name = module_map.get(input_type)
    if module_name:
        try:
            module = importlib.import_module(f".{module_name}", package=__name__)
            return getattr(module, module_name.capitalize())  # Capitalize to get class name
        except (ImportError, AttributeError) as e:
            raise ImportError(f"Could not import module or class for {input_type}: {e}") from e
    else:
        return None  # Or raise an appropriate error for unknown input_type



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"kapitan error classes"


class KapitanError(Exception):
    """generic kapitan error"""


class CompileError(KapitanError):
    """compile error"""


class InventoryError(KapitanError):
    """inventory error"""


class InventoryValidationError(InventoryError):
    """inventory validation error"""


class InvalidTargetError(InventoryError):
    """inventory validation error"""


class SecretError(KapitanError):
    """secrets error"""


class RefError(KapitanError):
    """ref error"""


class RefBackendError(KapitanError):
    """ref backend error"""


class RefFromFuncError(KapitanError):
    """ref from func error"""


class RefHashMismatchError(KapitanError):
    """ref has mismatch error"""


class HelmBindingUnavailableError(KapitanError):
    """helm input is used when the binding is not available"""


class HelmFetchingError(KapitanError):
    pass


class HelmTemplateError(KapitanError):
    pass


class GitSubdirNotFoundError(KapitanError):
    """git dependency subdir not found error"""


class GitFetchingError(KapitanError):
    """repo not found and/or permission error"""


class RequestUnsuccessfulError(KapitanError):
    """request error"""


class KubernetesManifestValidationError(KapitanError):
    """kubernetes manifest schema validation error"""



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"initialiser module"

import glob
import logging
import os

from copier import run_copy

from kapitan.errors import KapitanError
from kapitan.version import VERSION as kapitan_version

logger = logging.getLogger(__name__)


def initialise_skeleton(args):
    """Initialises a directory with a recommended skeleton structure using cruft
    Args:
        args.template_git_url (string): path or url that contains the cruft repository template
        args.checkout_ref (string): branch, tag or commit to checkout from the template repository
        args.directory (string): directory to initialise the skeleton in (default is current directory)
    """

    template_git_url = args.template_git_url
    checkout_ref = args.checkout_ref
    directory = os.path.abspath(args.directory)

    if set(glob.iglob(os.path.join(directory, "./*"))):
        raise KapitanError(f"Directory {directory} is not empty. Please initialise in an empty directory.")

    logger.info(f"Initialising kapitan from {template_git_url}@{checkout_ref} in {directory}")
    user_defaults = {
        "kapitan_version": kapitan_version,
    }
    run_copy(
        template_git_url,
        vcs_ref=args.checkout_ref,
        unsafe=True,
        dst_path=directory,
        quiet=True,
        user_defaults=user_defaults,
        skip_answered=False,
    )

    if directory == os.path.abspath(os.path.curdir):
        logger.info(f"Successfully initialised: run `kapitan --version`")
    else:
        logger.info(f"Successfully initialised kapitan in {directory}")
        logger.info("Please go to the directory and run `kapitan --version`")



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import re
import subprocess
from typing import Dict, List

from kapitan.inputs.base import InputType
from kapitan.inventory.model.input_types import KapitanInputTypeExternalConfig

logger = logging.getLogger(__name__)


class External(InputType):
    """
    External input type. Executes an external command to generate Kubernetes manifests.
    """

    env_vars: Dict[str, str] = {}  #: Environment variables to pass to the external command.
    command_args: List[str] = []  #: Command-line arguments to pass to the external command.

    def set_env_vars(self, env_vars):
        """
        Sets environment variables for the external command.
        Propagates HOME and PATH environment variables if they are not explicitly set.
        This is necessary because `subprocess.run()` with `env` set doesn't propagate
        environment variables from the current process.  We only propagate HOME or PATH if they
        exist in the Kapitan environment but aren't explicitly specified in `env_vars`. This
        prevents issues when spawning the subprocess due to `None` values in the subprocess's
        environment.
        """
        if "PATH" not in env_vars and "PATH" in os.environ:
            env_vars["PATH"] = os.environ["PATH"]
        if "HOME" not in env_vars and "HOME" in os.environ:
            env_vars["HOME"] = os.environ["HOME"]
        self.env_vars = env_vars

    def set_args(self, args: List[str]):
        """Sets command-line arguments for the external command."""
        self.command_args = args

    def compile_file(self, config: KapitanInputTypeExternalConfig, input_path, compile_path):
        """
        Execute external with specific arguments and env vars.
        If the external command exits with a non-zero error code, an error is raised.
        Args:
            config: KapitanInputTypeExternalConfig object.
            input_path: Path to the external command.
            compile_path: Path to the compiled target directory.
        """

        try:
            # file_path (str): Path to executable script or binary
            external_path = input_path

            self.set_args(config.args)
            self.set_env_vars(config.env_vars)

            args = [external_path]
            args.extend(self.command_args)
            args_str = " ".join(args)  # join args for logging and substitution

            # Substitute `${compiled_target_dir}` in the command arguments and environment variables.
            compiled_target_pattern = re.compile(r"(\${compiled_target_dir})")
            args_str = compiled_target_pattern.sub(compile_path, args_str)
            env_vars = {k: compiled_target_pattern.sub(compile_path, v) for (k, v) in config.env_vars.items()}

            # Run the external command.  shell=True is required for argument substitution to work correctly.
            # However, this introduces a security risk if the input_path or command_args are not properly sanitized.
            # Consider using the shlex module to properly quote and escape arguments to mitigate this risk.
            # See https://docs.python.org/3/library/shlex.html for more information.

            logger.debug("Executing external input with command '%s' and env vars '%s'.", args, env_vars)

            external_result = subprocess.run(
                args_str,
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                encoding="utf-8",
            )

            logger.debug("External stdout: %s.", external_result.stdout)
            if external_result.returncode != 0:
                raise ValueError(
                    "Executing external input with command '{}' and env vars '{}' failed: {}".format(
                        args, env_vars, external_result.stderr
                    )
                )

        except OSError as e:
            logger.exception("External failed to run. Error: %s", e)



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"hashicorp vault transit secrets module"

import base64
import logging
from binascii import Error as b_error
from sys import exit

from hvac.exceptions import Forbidden, InvalidPath

from kapitan import cached
from kapitan.inventory.model.references import KapitanReferenceVaultTransitConfig
from kapitan.refs import KapitanReferencesTypes
from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend
from kapitan.refs.vault_resources import VaultClient, VaultError

logger = logging.getLogger(__name__)


class VaultTransit(Base64Ref):
    """
    Hashicorp Vault support for Transit Secret Engine
    """

    def __init__(
        self,
        data,
        vault_params: KapitanReferenceVaultTransitConfig,
        encrypt=True,
        encode_base64=False,
        **kwargs,
    ):
        """
        Set vault parameter and encoding of data
        """
        if isinstance(vault_params, dict):
            vault_params = KapitanReferenceVaultTransitConfig(**vault_params)
        self.vault_params = vault_params

        if encrypt:
            self._encrypt(data, self.vault_params.crypto_key, True)
        else:
            self.data = data

        super().__init__(self.data, **kwargs)
        self.type_name = KapitanReferencesTypes.VAULTTRANSIT

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new VaultSecret from data and ref_params: target_name
        parameters will be grabbed from the inventory via target_name
        """
        try:
            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv.get_parameters(target_name)

            ref_params.kwargs["vault_params"] = target_inv.kapitan.secrets.vaulttransit

            return cls(data, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create VaultSecret: vaulttransit parameters missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        Returns decrypted data
        """
        # can't use super().reveal() as we want bytes
        try:
            ref_data = base64.b64decode(self.data, validate=True)
        except b_error:
            exit("non-alphabet characters in the data")

        return self._decrypt(ref_data)

    def update_key(self, key):
        """
        re-encrypts data with new key, respects original encoding
        returns True if key is different and secret is updated, False otherwise
        """
        if key == self.vault_params.crypto_key:
            return False

        data_dec = self.reveal()

        encode_base64 = self.encoding == "base64"
        if encode_base64:
            logger.info('Content is already base64 encoded "?(vaulttransit:...|base64)" has no effect.')
            data_dec = base64.b64decode(data_dec).decode()
        self._encrypt(data_dec, key, encode_base64)
        self.data = base64.b64encode(self.data).decode()

        return True

    def _encrypt(self, data, key, encode_base64):
        """
        Authenticate with Vault server & returns value of the key from secret

        :returns: secret in plain text
        """
        _data = data
        self.encoding = "original"
        if not encode_base64:
            _data = base64.b64encode(data.encode())
            self.encoding = "base64"
        else:
            # To guarantee _data is bytes
            if isinstance(data, str):
                _data = data.encode()

        client = VaultClient(self.vault_params)
        # token will comprise of two parts path_in_vault:key
        # data = self.data.decode("utf-8").rstrip().split(":")

        try:
            # Request encryption by vault
            response = client.secrets.transit.encrypt_data(
                name=key,
                mount_point=self.vault_params.mount,
                plaintext=base64.b64encode(_data).decode("ascii"),
            )
            ciphertext = response["data"]["ciphertext"]

            self.data = ciphertext.encode()
            self.vault_params.crypto_key = key
        except Forbidden:
            raise VaultError(
                "Permission Denied. "
                + "make sure the token is authorised to access {path} on Vault".format(path=data[0])
            )
        except InvalidPath:
            raise VaultError("{path} does not exist on Vault secret".format(path=data[0]))
        finally:
            client.adapter.close()

    def _decrypt(self, data):
        """
        Authenticate with Vault server & returns value of the key from secret

        :returns: secret in plain text
        """
        client = VaultClient(self.vault_params)
        always_latest = self.vault_params.always_latest
        key = self.vault_params.crypto_key
        if key is None:
            raise RefError("Cannot access vault params")

        try:
            if always_latest:
                encrypt_data_response = client.secrets.transit.rewrap_data(
                    name=key,
                    mount_point=self.vault_params.mount,
                    ciphertext=data.decode(),
                )
                data = encrypt_data_response["data"]["ciphertext"].encode()

            response = client.secrets.transit.decrypt_data(
                name=key,
                mount_point=self.vault_params.mount,
                ciphertext=data.decode(),
            )
            plaintext = base64.b64decode(response["data"]["plaintext"])

            return plaintext.decode()
        except Forbidden:
            raise VaultError(
                "Permission Denied. "
                + "make sure the token is authorised to access {path} on Vault".format(path=data[0])
            )
        except InvalidPath:
            raise VaultError("{path} does not exist on Vault secret".format(path=data[0]))
        finally:
            client.adapter.close()

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {
            "data": self.data,
            "encoding": self.encoding,
            "type": self.type_name,
            "vault_params": self.vault_params.model_dump(),
        }


class VaultBackend(Base64RefBackend):
    def __init__(self, path, ref_type=VaultTransit, **ref_kwargs):
        "init VaultBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = KapitanReferencesTypes.VAULTTRANSIT



import os
import sys

from regex import regex

from kapitan.inventory.inventory import InventoryError


def migrate(inventory_path: str = "inventory"):
    # FEAT: write migrations to temp dir and copy only if succeeded

    if os.path.exists(inventory_path):
        if os.path.isdir(inventory_path):
            migrate_dir(inventory_path)
        elif os.path.isfile(inventory_path):
            migrate_file(inventory_path)
    else:
        print(f"Error while migrating: inventory path at {inventory_path} does not exist")


def migrate_dir(path: str):
    """migrates all .yml/.yaml files in the given path to omegaconfs syntax"""

    for root, _, files in os.walk(path):
        for file in files:
            file = os.path.join(root, file)
            name, ext = os.path.splitext(file)

            if ext not in (".yml", ".yaml"):
                continue

            try:
                migrate_file(file)
            except Exception as e:
                InventoryError(f"{file}: error with migration: {e}")


def migrate_file(file: str):
    with open(file, "r") as fp:
        content = fp.read()

    updated_content = migrate_str(content)

    with open(file, "w") as fp:
        fp.write(updated_content)


def migrate_str(content: str):
    # FEAT: don't migrate custom resolvers
    # FEAT: migrate interpolations with '.' in the keyname

    # search for interpolation pattern
    # migrate path delimiter
    # migrate meta data name
    updated_content = regex.sub(
        r"(?<!\\)\${([^\${}]*+(?:(?R)[^\${}]*)*+)}",
        lambda match: "${"
        + match.group(1)
        .replace(
            ":",
            ".",
        )
        .replace("_reclass_", "_kapitan_")
        + "}",
        content,
    )

    # replace escaped tags with specific resolver
    excluded_chars = "!()"
    invalid = any(c in updated_content for c in excluded_chars)
    updated_content = regex.sub(
        r"\\\${([^\${}]*+(?:(?R)[^\${}]*)*+)}",
        lambda match: ("${escape:" if not invalid else "\\\\\\${") + match.group(1) + "}",
        updated_content,
    )

    return updated_content


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: ./migrate.py <inventory-path>")
        sys.exit(1)

    print(f"Migrating all .yml/.yaml files in {sys.argv[1]}")
    migrate(sys.argv[1])



from __future__ import print_function

import base64
import logging
import mimetypes
import os
import re
import sys

from kapitan.errors import KapitanError, RefError, RefHashMismatchError
from kapitan.inventory.model.references import (
    KapitanReferenceConfig,
    KapitanReferenceVaultKVConfig,
    KapitanReferenceVaultTransitConfig,
)
from kapitan.refs import KapitanReferencesTypes
from kapitan.refs.base import PlainRef, RefController, Revealer
from kapitan.refs.base64 import Base64Ref
from kapitan.refs.env import EnvRef
from kapitan.refs.secrets.awskms import AWSKMSSecret
from kapitan.refs.secrets.azkms import AzureKMSSecret
from kapitan.refs.secrets.gkms import GoogleKMSSecret
from kapitan.refs.secrets.gpg import GPGSecret, lookup_fingerprints
from kapitan.refs.secrets.vaultkv import VaultSecret
from kapitan.refs.secrets.vaulttransit import VaultTransit
from kapitan.resources import get_inventory
from kapitan.utils import fatal_error, search_target_token_paths

logger = logging.getLogger(__name__)


def handle_refs_command(args):
    ref_controller = RefController(args.refs_path)

    if args.write is not None:
        ref_write(args, ref_controller)
    elif args.reveal:
        ref_reveal(args, ref_controller)
    elif args.update:
        secret_update(args, ref_controller)
    elif args.update_targets or args.validate_targets:
        secret_update_validate(args, ref_controller)


def ref_write(args, ref_controller):
    "Write ref to ref_controller based on cli args"
    token_name = args.write
    file_name = args.file
    is_binary = args.binary
    data = None

    if file_name is None:
        fatal_error("--file is required with --write")
    if file_name == "-":
        data = ""
        for line in sys.stdin:
            data += line
    else:
        mimetypes.guess_type(file_name)
        modifier = "rb" if is_binary else "r"
        with open(file_name, modifier) as fp:
            try:
                data = fp.read()
            except UnicodeDecodeError as e:
                raise KapitanError(
                    "Could not read file. Please add '--binary' if the file contains binary data. ({})".format(
                        e
                    )
                )

    # Empty configuration object
    reference_backend_configs = KapitanReferenceConfig()

    if args.target_name:
        inv = get_inventory(args.inventory_path)

        try:
            reference_backend_configs = inv.get_parameters(args.target_name).kapitan.secrets
        except (KeyError, AttributeError):
            raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

    type_name, token_path = token_name.split(":")

    try:
        type_name = KapitanReferencesTypes(type_name)
    except ValueError:
        raise KapitanError(
            f"Invalid token type: {type_name}. Try using gpg/gkms/awskms/azkms/vaultkv/vaulttransit/base64/plain/env"
        )

    tag = f"?{{{type_name}:{token_path}}}"

    if type_name == KapitanReferencesTypes.GPG:
        # args.recipients is a list, convert to recipients dict
        recipients = [dict((("name", name),)) for name in args.recipients]

        if reference_backend_configs.gpg:
            recipients = reference_backend_configs.gpg.recipients

        if not recipients:
            raise KapitanError(
                "No GPG recipients specified. Use --recipients or specify them in "
                + "parameters.kapitan.secrets.gpg.recipients and use --target"
            )

        secret_obj = GPGSecret(data, recipients, encode_base64=args.base64)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.GKMS:
        key = args.key

        if reference_backend_configs.gkms:
            key = args.key or reference_backend_configs.gkms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target"
            )

        logger.debug(f"Using gkms key {key}")

        secret_obj = GoogleKMSSecret(data, key, encode_base64=args.base64)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.AWSKMS:
        key = args.key

        if reference_backend_configs.awskms:
            key = args.key or reference_backend_configs.awskms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target"
            )

        logger.debug(f"Using awskms key {key}")
        secret_obj = AWSKMSSecret(data, key, encode_base64=args.base64)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.AZKMS:
        key = args.key

        if reference_backend_configs.azkms:
            key = args.key or reference_backend_configs.azkms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.azkms.key and use --target"
            )

        logger.debug(f"Using azkms key {key}")
        secret_obj = AzureKMSSecret(data, key, encode_base64=args.base64)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.BASE64:
        _data = data if is_binary else data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = Base64Ref(_data, encoding=encoding)
        ref_controller[tag] = ref_obj

    # VAULT Key-Value Engine
    elif type_name == KapitanReferencesTypes.VAULTKV:
        _data = data if is_binary else data.encode()
        encoding = "original"

        vault_params = KapitanReferenceVaultKVConfig()

        if reference_backend_configs.vaultkv:
            vault_params = reference_backend_configs.vaultkv

        if args.vault_auth:
            vault_params.auth = args.vault_auth
        if not vault_params.auth:
            raise KapitanError(
                "No Authentication type parameter specified. Specify it"
                " in parameters.kapitan.secrets.vaultkv.auth and use --target-name or use --vault-auth"
            )

        kwargs = {}

        # set mount
        mount = args.vault_mount
        if vault_params.mount:
            mount = vault_params.mount

        kwargs["mount_in_vault"] = mount

        # set path in vault
        path_in_vault = args.vault_path
        if not path_in_vault:
            path_in_vault = token_path  # token path in kapitan as default
        kwargs["path_in_vault"] = path_in_vault

        # set key
        key = args.vault_key
        if key:
            kwargs["key_in_vault"] = key
        else:
            raise RefError("Could not create VaultSecret: vaultkv: key is missing")

        secret_obj = VaultSecret(_data, vault_params, **kwargs)
        ref_controller[tag] = secret_obj

    # VAULT Transit engine
    elif type_name == KapitanReferencesTypes.VAULTTRANSIT:
        _data = data.encode()
        vault_params = KapitanReferenceVaultTransitConfig()

        if reference_backend_configs.vaulttransit:
            vault_params = reference_backend_configs.vaulttransit

        if args.vault_auth:
            vault_params.auth = args.vault_auth

        if not vault_params.auth:
            raise KapitanError(
                "No Authentication type parameter specified. Specify it"
                " in parameters.kapitan.secrets.vaultkv.auth and use --target-name or use --vault-auth"
            )

        secret_obj = VaultTransit(_data, vault_params)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.PLAIN:
        _data = data if is_binary else data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = PlainRef(_data, encoding=encoding)
        ref_controller[tag] = ref_obj

    elif type_name == KapitanReferencesTypes.ENV:
        _data = data if is_binary else data.encode()
        encoding = "original"
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = "base64"
        ref_obj = EnvRef(_data, encoding=encoding)
        ref_controller[tag] = ref_obj


def secret_update(args, ref_controller):
    "Update secret gpg recipients/gkms/awskms key"
    # TODO --update *might* mean something else for other types
    token_name = args.update
    reference_backend_configs = KapitanReferenceConfig()

    if args.target_name:
        inv = get_inventory(args.inventory_path)

        if not inv.get_parameters(args.target_name).kapitan.secrets:
            reference_backend_configs = inv.get_parameters(args.target_name).kapitan.secrets
            raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

    type_name, token_path = token_name.split(":")

    if type_name not in KapitanReferencesTypes:
        raise KapitanError(
            f"Invalid token type: {type_name}. Try using gpg/gkms/awskms/azkms/vaultkv/vaulttransit/base64/plain/env"
        )

    tag = f"?{{{type_name}:{token_path}}}"

    if type_name == KapitanReferencesTypes.GPG:
        # args.recipients is a list, convert to recipients dict
        recipients = [dict((("name", name),)) for name in args.recipients]

        if reference_backend_configs.gpg:
            recipients = reference_backend_configs.gpg.recipients

        if not recipients:
            raise KapitanError(
                "No GPG recipients specified. Use --recipients or specify them in "
                + "parameters.kapitan.secrets.gpg.recipients and use --target"
            )

        secret_obj = ref_controller[tag]
        secret_obj.update_recipients(recipients)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.GKMS:
        key = args.key

        if reference_backend_configs.gkms:
            key = reference_backend_configs.gkms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target"
            )
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.AZKMS:
        key = args.key

        if reference_backend_configs.azkms:
            key = reference_backend_configs.azkms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.azkms.key and use --target"
            )
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    elif type_name == KapitanReferencesTypes.AWSKMS:
        key = args.key

        if reference_backend_configs.awskms:
            key = reference_backend_configs.awskms.key

        if not key:
            raise KapitanError(
                "No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target"
            )

        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    else:
        fatal_error("Invalid token: {name}. Try using gpg/gkms/awskms:{name}".format(name=token_name))


def ref_reveal(args, ref_controller):
    "Reveal secrets in file_name"
    revealer = Revealer(ref_controller)
    file_name = args.file
    reffile_name = args.ref_file
    tag_name = args.tag

    if file_name is None and reffile_name is None and tag_name is None:
        fatal_error("--file or --ref-file is required with --reveal")
    try:
        if file_name == "-" or reffile_name == "-":
            out = revealer.reveal_raw_file(None)
            sys.stdout.write(out)
        elif file_name:
            for rev_obj in revealer.reveal_path(file_name):
                sys.stdout.write(rev_obj.content)
        elif reffile_name:
            ref = ref_controller.ref_from_ref_file(reffile_name)
            sys.stdout.write(ref.reveal())
        elif tag_name:
            out = revealer.reveal_raw_string(tag_name)
            sys.stdout.write(out)
    except (RefHashMismatchError, KeyError):
        raise KapitanError("Reveal failed for file {name}".format(name=file_name))


def secret_update_validate(args, ref_controller):
    "Validate and/or update target secrets"
    # update gpg recipients/gkms/awskms key for all secrets in secrets_path
    # use --refs-path to set scanning path
    inv = get_inventory(args.inventory_path)
    targets = set(inv.targets.keys())
    secrets_path = os.path.abspath(args.refs_path)
    target_token_paths = search_target_token_paths(secrets_path, targets)
    ret_code = 0

    for target_name, token_paths in target_token_paths.items():
        secrets = inv.get_parameters(target_name).kapitan.secrets
        if not secrets:
            raise KapitanError("parameters.kapitan.secrets not defined in {}".format(target_name))

        for token_path in token_paths:
            type_name = re.match(r"\?\{(\w+):", token_path).group(1)
            if type_name == KapitanReferencesTypes.GPG:
                if not secrets.gpg:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory gpg recipients, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                recipients = secrets.gpg.recipients
                secret_obj = ref_controller[token_path]
                target_fingerprints = set(lookup_fingerprints(recipients))
                secret_fingerprints = set(lookup_fingerprints(secret_obj.recipients))
                if target_fingerprints != secret_fingerprints:
                    if args.validate_targets:
                        logger.info("%s recipient mismatch", token_path)
                        to_remove = secret_fingerprints.difference(target_fingerprints)
                        to_add = target_fingerprints.difference(secret_fingerprints)
                        if to_remove:
                            logger.info("%s needs removal", to_remove)
                        if to_add:
                            logger.info("%s needs addition", to_add)
                        ret_code = 1
                    else:
                        new_recipients = [
                            dict(
                                [
                                    ("fingerprint", f),
                                ]
                            )
                            for f in target_fingerprints
                        ]
                        secret_obj.update_recipients(new_recipients)
                        ref_controller[token_path] = secret_obj

            elif type_name == KapitanReferencesTypes.GKMS:
                if not secrets.gkms:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory gkms key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                key = secrets.gkms.key
                secret_obj = ref_controller[token_path]
                if secrets.gpg.key != key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(key)
                        ref_controller[token_path] = secret_obj

            elif type_name == KapitanReferencesTypes.VAULTTRANSIT:
                if not secrets.vaulttransit:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory vaulttransit key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                secret_obj = ref_controller[token_path]
                key = secrets.vaulttransit.key
                if key != secret_obj.vault_params["key"]:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(key)
                        ref_controller[token_path] = secret_obj

            elif type_name == KapitanReferencesTypes.AWSKMS:
                if not secrets.awskms:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory awskms key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                key = secrets.awskms.key
                secret_obj = ref_controller[token_path]
                if key != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(key)
                        ref_controller[token_path] = secret_obj

            elif type_name == KapitanReferencesTypes.AZKMS:
                if not secrets.azkey:
                    logger.debug(
                        "secret_update_validate: target: %s has no inventory azkms key, skipping %s",
                        target_name,
                        token_path,
                    )
                    continue
                secret_obj = ref_controller[token_path]
                key = secrets.azkms.key
                if key != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(key)
                        ref_controller[token_path] = secret_obj

            else:
                logger.info("Invalid secret %s, could not get type, skipping", token_path)

    sys.exit(ret_code)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import shutil

from kapitan.inputs.base import InputType
from kapitan.inventory.model.input_types import KapitanInputTypeCopyConfig

logger = logging.getLogger(__name__)


class Remove(InputType):

    def compile_file(self, config: KapitanInputTypeCopyConfig, input_path, compile_path):
        """Remove a file or directory.

        Args:
            config (KapitanInputTypeCopyConfig): input type configuration
            input_path (str): path to file or directory to remove
            compile_path (str): not used in this input type

        Raises:
            OSError: if the file or directory cannot be removed
        """

        try:
            logger.debug("Removing %s", input_path)
            if os.path.isfile(input_path):
                os.remove(input_path)
            else:
                shutil.rmtree(input_path)
        except OSError as e:
            logger.exception("Input dir not removed. Error: %s", e)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"kapitan targets"
import logging
import multiprocessing
import os
import shutil
import tempfile
import time
from functools import partial

from reclass.errors import NotFoundError, ReclassException

from kapitan import cached
from kapitan.dependency_manager.base import fetch_dependencies
from kapitan.errors import CompileError, InventoryError, KapitanError
from kapitan.inputs import get_compiler
from kapitan.resources import get_inventory

logger = logging.getLogger(__name__)


def compile_targets(inventory_path, search_paths, ref_controller, args):
    """
    Searches and loads target files, and runs compile_target() on a
    multiprocessing pool with parallel number of processes.
    kwargs are passed to compile_target()
    """
    # temp_path will hold compiled items
    temp_path = tempfile.mkdtemp(suffix=".kapitan")
    # enable previously compiled items to be reference in other compile inputs
    search_paths.append(temp_path)
    temp_compile_path = os.path.join(temp_path, "compiled")
    dep_cache_dir = temp_path

    rendering_start = time.time()
    inventory = get_inventory(inventory_path)
    discovered_targets = inventory.targets.keys()

    logger.info(
        f"Rendered inventory (%.2fs): discovered {len(discovered_targets)} targets.",
        time.time() - rendering_start,
    )

    if discovered_targets == 0:
        raise CompileError("No inventory targets discovered at path: {inventory_path}")

    targets = args.targets or discovered_targets
    labels = args.labels

    try:
        targets = search_targets(inventory, targets, labels)

    except CompileError as e:
        raise CompileError(f"Error searching targets: {e}")

    if len(targets) == 0:
        raise CompileError(f"No matching targets found in inventory: {labels if labels else args.targets}")

    parallelism = args.parallelism or min(len(targets), os.cpu_count())

    logger.info(
        f"Compiling {len(targets)}/{len(discovered_targets)} targets using {parallelism} concurrent processes: ({os.cpu_count()} CPU detected)"
    )

    with multiprocessing.Pool(parallelism) as pool:
        try:
            fetching_start = time.time()
            # check if --fetch or --force-fetch is enabled
            force_fetch = args.force_fetch
            fetch = args.fetch or force_fetch

            # deprecated --force flag
            if args.force:
                logger.info(
                    "DeprecationWarning: --force is deprecated. Use --force-fetch instead of --force --fetch"
                )
                force_fetch = True

            if fetch:
                # skip classes that are not yet available
                target_objs = load_target_inventory(inventory, targets, ignore_class_not_found=True)
            else:
                # ignore_class_not_found = False by default
                target_objs = load_target_inventory(inventory, targets)

            # append "compiled" to output_path so we can safely overwrite it
            output_path = args.output_path
            compile_path = os.path.join(output_path, "compiled")

            if not target_objs:
                raise CompileError("Error: no targets found")

            # fetch dependencies
            if fetch:
                fetch_dependencies(output_path, target_objs, dep_cache_dir, force_fetch, pool)
            # fetch targets which have force_fetch: true
            elif not force_fetch:
                fetch_objs = []
                # iterate through targets
                for target in target_objs:
                    for entry in target.dependencies:
                        force_fetch = entry.force_fetch
                        if entry.force_fetch:
                            fetch_objs.append(target)

                # fetch dependencies from targets with force_fetch set to true
                if fetch_objs:
                    fetch_dependencies(output_path, fetch_objs, dep_cache_dir, True, pool)
                    logger.info("Fetched dependencies (%.2fs)", time.time() - fetching_start)

            compile_start = time.time()
            worker = partial(
                compile_target,
                search_paths=search_paths,
                compile_path=temp_compile_path,
                ref_controller=ref_controller,
                globals_cached=cached.as_dict(),
                args=args,
            )

            # compile_target() returns None on success
            # so p is only not None when raising an exception
            [p.get() for p in pool.imap_unordered(worker, target_objs) if p]

            os.makedirs(compile_path, exist_ok=True)

            # if '-t' is set on compile or only a few changed, only override selected targets
            if len(target_objs) < len(discovered_targets):
                for target in target_objs:
                    path = target.target_full_path
                    compile_path_target = os.path.join(compile_path, path)
                    temp_path_target = os.path.join(temp_compile_path, path)

                    os.makedirs(compile_path_target, exist_ok=True)

                    shutil.rmtree(compile_path_target)
                    shutil.copytree(temp_path_target, compile_path_target)
                    logger.debug("Copied %s into %s", temp_path_target, compile_path_target)
            # otherwise override all targets
            else:
                shutil.rmtree(compile_path)
                shutil.copytree(temp_compile_path, compile_path)
                logger.debug("Copied %s into %s", temp_compile_path, compile_path)
            logger.info(f"Compiled {len(targets)} targets in %.2fs", time.time() - compile_start)
        except ReclassException as e:
            if isinstance(e, NotFoundError):
                logger.error("Inventory reclass error: inventory not found")
            else:
                logger.error("Inventory reclass error: %s", e.message)
            raise InventoryError(e.message)
        except Exception as e:
            # if compile worker fails, terminate immediately
            pool.terminate()
            logger.debug("Compile pool terminated")
            # only print traceback for errors we don't know about
            if not isinstance(e, KapitanError):
                logger.exception("\nUnknown (Non-Kapitan) error occurred:\n")

            logger.error("\n")
            if args.verbose:
                logger.exception(e)
            else:
                logger.error(e)
            raise CompileError(f"Error compiling targets: {e}")

        finally:
            shutil.rmtree(temp_path)
            logger.debug("Removed %s", temp_path)


def load_target_inventory(inventory, requested_targets, ignore_class_not_found=False):
    """returns a list of target objects from the inventory"""
    target_objs = []

    # if '-t' is set on compile, only loop through selected targets
    if requested_targets:
        targets = inventory.get_targets(requested_targets)
    else:
        targets = inventory.targets

    for target_name, target in targets.items():
        try:
            if not target.parameters:
                if ignore_class_not_found:
                    continue
                else:
                    raise InventoryError(f"InventoryError: {target_name}: parameters is empty")

            kapitan_target_configs = target.parameters.kapitan
            # check if parameters.kapitan is empty
            if not kapitan_target_configs:
                raise InventoryError(f"InventoryError: {target_name}: parameters.kapitan has no assignment")
            kapitan_target_configs.target_full_path = inventory.targets[target_name].name.replace(".", "/")
            logger.debug(f"load_target_inventory: found valid kapitan target {target_name}")
            target_objs.append(kapitan_target_configs)
        except KeyError:
            logger.debug(f"load_target_inventory: target {target_name} has no kapitan compile obj")

    return target_objs


def search_targets(inventory, targets, labels):
    """returns a list of targets where the labels match, otherwise just return the original targets"""
    if not labels:
        return targets

    try:
        labels_dict = dict(label.split("=") for label in labels)
    except ValueError:
        raise CompileError(
            "Compile error: Failed to parse labels, should be formatted like: kapitan compile -l env=prod app=example"
        )

    targets_found = []
    # It should come back already rendered

    for target in inventory.targets.values():
        target_labels = target.parameters.kapitan.labels
        matched_all_labels = False
        for label, value in labels_dict.items():
            try:
                if target_labels[label] == value:
                    matched_all_labels = True
                    continue
            except KeyError:
                logger.debug(
                    f"search_targets: label {label}={value} didn't match target {target.name} {target_labels}"
                )

            matched_all_labels = False
            break

        if matched_all_labels:
            targets_found.append(target.name)

    if len(targets_found) == 0:
        raise CompileError(f"No targets found with labels: {labels}")

    return targets_found


def compile_target(target_config, search_paths, compile_path, ref_controller, args, globals_cached=None):
    """Compiles target_obj and writes to compile_path"""
    start = time.time()
    compile_configs = target_config.compile
    target_name = target_config.vars.target

    # Only populates the cache if the subprocess doesn't have it
    if globals_cached and not cached.inv:
        cached.from_dict(globals_cached)

    for compile_config in compile_configs:
        try:
            input_type = compile_config.input_type
            input_compiler = get_compiler(input_type)(
                compile_path, search_paths, ref_controller, target_name, args
            )
            input_compiler.compile_obj(compile_config)
        except AttributeError as e:
            import traceback

            traceback.print_exception(type(e), e, e.__traceback__)
            raise CompileError(f'Invalid input_type: "{compile_config.input_type}" {e}') from e

        except Exception as e:
            if compile_config.continue_on_compile_error:
                logger.error("Error compiling %s: %s", target_name, e)
                continue
            else:
                import traceback

                traceback.print_exception(type(e), e, e.__traceback__)
                raise CompileError(f"Error compiling {target_name}: {e}") from e

    logger.info("Compiled %s (%.2fs)", target_config.target_full_path, time.time() - start)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import tempfile

import yaml

from kapitan.errors import HelmTemplateError
from kapitan.helm_cli import helm_cli
from kapitan.inputs.base import InputType
from kapitan.inputs.kadet import BaseModel, BaseObj
from kapitan.inventory.model.input_types import KapitanInputTypeHelmConfig

logger = logging.getLogger(__name__)

HELM_DENIED_FLAGS = {
    "dry-run",
    "generate-name",
    "help",
    "output-dir",
    "show-only",
}

HELM_DEFAULT_FLAGS = {"--include-crds": True, "--skip-tests": True}


class Helm(InputType):
    def compile_file(self, config: KapitanInputTypeHelmConfig, input_path, compile_path):
        """Render templates in input_path/templates and write to compile_path.
        input_path must be a directory containing a helm chart.

        kwargs:
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled

        Raises:
            HelmTemplateError: if helm template fails

        """
        helm_values_files = config.helm_values_files
        helm_params = config.helm_params
        helm_path = config.helm_path

        helm_values_file = None
        if config.helm_values:
            helm_values_file = write_helm_values_file(config.helm_values)

        helm_flags = dict(HELM_DEFAULT_FLAGS)
        # add to flags if set
        if config.kube_version:
            helm_flags["--api-versions"] = config.kube_version

        temp_dir = tempfile.mkdtemp()
        # save the template output to temp dir first
        _, error_message = render_chart(
            chart_dir=input_path,
            output_path=temp_dir,
            helm_path=helm_path,
            helm_params=helm_params,
            helm_values_file=helm_values_file,
            helm_values_files=helm_values_files,
            helm_flags=helm_flags,
        )
        if error_message:
            raise HelmTemplateError(error_message)
        # Iterate over all files in the temporary directory

        walk_root_files = os.walk(temp_dir)
        for current_dir, _, files in walk_root_files:
            for file in files:  # go through all the template files
                rel_dir = os.path.relpath(current_dir, temp_dir)
                rel_file_name = os.path.join(rel_dir, file)
                full_file_name = os.path.join(current_dir, file)
                # Open each file and write its content to the compilation path
                with open(full_file_name, "r", encoding="utf-8") as f:
                    file_path = os.path.join(compile_path, rel_file_name)
                    os.makedirs(os.path.dirname(file_path), exist_ok=True)
                    item_value = list(yaml.safe_load_all(f))
                    self.to_file(config, file_path, item_value)

    def render_chart(self, *args, **kwargs):
        return render_chart(*args, **kwargs)


def render_chart(
    chart_dir,
    output_path,
    helm_path,
    helm_params,
    helm_values_file,
    helm_values_files,
    helm_flags=None,
):
    """
    Renders helm chart located at chart_dir.

    Args:
        output_path: path to write rendered chart. If '-', returns rendered chart as string.

    Returns:
        tuple: (output, error_message)
    """
    args = ["template"]

    name = helm_params.pop("name", None)
    output_file = helm_params.pop("output_file", None)

    if helm_flags is None:
        helm_flags = HELM_DEFAULT_FLAGS

    # Validate and process helm parameters
    for param, value in helm_params.items():
        if len(param) == 1:
            raise ValueError(f"invalid helm flag: '{param}'. helm_params supports only long flag names")

        if "-" in param:
            raise ValueError(f"helm flag names must use '_' and not '-': {param}")

        param = param.replace("_", "-")

        if param in ("set", "set-file", "set-string"):
            raise ValueError(
                f"helm '{param}' flag is not supported. Use 'helm_values' to specify template values"
            )

        if param == "values":
            raise ValueError(
                f"helm '{param}' flag is not supported. Use 'helm_values_files' to specify template values files"
            )

        if param in HELM_DENIED_FLAGS:
            raise ValueError(f"helm flag '{param}' is not supported.")

        # Set helm flags
        helm_flags[f"--{param}"] = value

    # 'release_name' used to be the "helm template" [NAME] parameter.
    # For backward compatibility, assume it is the '--release-name' flag only if its value is a bool.
    release_name = helm_flags.get("--release-name")
    if release_name is not None and not isinstance(release_name, bool):
        logger.warning("using 'release_name' to specify the output name is deprecated. Use 'name' instead")
        del helm_flags["--release-name"]
        # name is used in place of release_name if both are specified
        name = name or release_name

    # Add flags to args list
    for flag, value in helm_flags.items():
        # boolean flag should be passed when present, and omitted when not specified
        if isinstance(value, bool):
            if value:
                args.append(flag)
        else:
            args.append(flag)
            args.append(str(value))

    # Add values files to args list
    if helm_values_file:
        args.append("--values")
        args.append(helm_values_file)

    if helm_values_files:
        for file_name in helm_values_files:
            args.append("--values")
            args.append(file_name)

    # Set output directory
    if not output_file and output_path not in (None, "-"):
        args.append("--output-dir")
        args.append(output_path)

    if "name_template" not in helm_flags:
        args.append(name or "--generate-name")

    # Add chart directory to args list
    # uses absolute path to make sure helm interprets it as a
    # local dir and not a chart_name that it should download.
    args.append(chart_dir)

    # If output_path is '-', output is a string with rendered chart
    if output_path == "-":
        _, helm_output = tempfile.mkstemp(".helm_output.yml", text=True)
        with open(helm_output, "w+") as f:
            error_message = helm_cli(helm_path, args, stdout=f)
            f.seek(0)
            return (f.read(), error_message)

    if output_file:
        with open(os.path.join(output_path, output_file), "wb") as f:
            # can't be verbose when capturing stdout
            error_message = helm_cli(helm_path, args, stdout=f)
            return ("", error_message)
    else:
        error_message = helm_cli(helm_path, args, verbose="--debug" in helm_flags)
        return ("", error_message)


def write_helm_values_file(helm_values: dict):
    """Dump helm values into a temporary YAML file.

    Args:
        helm_values: A dictionary containing helm values.

    Returns:
        str: The path to the temporary YAML file.
    """
    _, helm_values_file = tempfile.mkstemp(".helm_values.yml", text=True)
    with open(helm_values_file, "w") as fp:
        yaml.safe_dump(helm_values, fp)

    return helm_values_file


class HelmChart(BaseModel):
    """
    Represents a Helm chart. Renders the chart and stores the rendered objects in self.root.

    Args:
        chart_dir: Path to the Helm chart directory.

    Raises:
        HelmTemplateError: if helm template fails
    """

    chart_dir: str
    helm_params: dict = {}
    helm_values: dict = {}
    helm_path: str = None

    # Load and process the Helm chart
    def new(self):
        for obj in self.load_chart():
            self.root[f"{obj['metadata']['name'].lower()}-{obj['kind'].lower().replace(':','-')}"] = (
                BaseObj.from_dict(obj)
            )

    def load_chart(self):
        helm_values_file = None
        if self.helm_values != {}:
            helm_values_file = write_helm_values_file(self.helm_values)
        output, error_message = render_chart(
            self.chart_dir, "-", self.helm_path, self.helm_params, helm_values_file, None
        )
        if error_message:
            raise HelmTemplateError(error_message)

        return yaml.safe_load_all(output)
